# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class RebuildGhostscriptTestCase(TestCase):
    def testRebuildGhostscript(self):
        """
        Test bug 703676, where app-text/libspectre was rebuilt before
        its app-text/ghostscript-gpl DEPEND.
        """
        ebuilds = {
            "app-text/ghostscript-gpl-10.01.1": {
                "EAPI": "8",
                "DEPEND": "gtk? ( x11-libs/gtk+:3 )",
                "RDEPEND": "gtk? ( x11-libs/gtk+:3 )",
                "IUSE": "gtk",
            },
            "app-text/ghostscript-gpl-10.01.2": {
                "EAPI": "8",
                "SLOT": "0/10.01",
                "DEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "RDEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "IUSE": "dbus gtk",
            },
            "app-text/libspectre-0.2.11": {
                "EAPI": "8",
                "DEPEND": ">=app-text/ghostscript-gpl-9.53.0:=",
                "RDEPEND": ">=app-text/ghostscript-gpl-9.53.0:=",
            },
            "app-text/libspectre-0.2.12": {
                "EAPI": "8",
                "DEPEND": ">=app-text/ghostscript-gpl-9.53.0:=",
                "RDEPEND": ">=app-text/ghostscript-gpl-9.53.0:=",
            },
            "net-dns/avahi-0.8-r7": {
                "EAPI": "8",
                "DEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "RDEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "IUSE": "dbus gtk",
            },
            "net-print/cups-2.4.6": {
                "EAPI": "8",
                "DEPEND": "zeroconf? ( >=net-dns/avahi-0.6.31-r2[dbus] )",
                "RDEPEND": "zeroconf? ( >=net-dns/avahi-0.6.31-r2[dbus] )",
                "IUSE": "zeroconf",
            },
            "sys-apps/dbus-1.15.6": {
                "EAPI": "8",
            },
            "x11-libs/gtk+-3.24.38": {
                "EAPI": "8",
                "SLOT": "3",
                "DEPEND": "cups? ( >=net-print/cups-2.0 )",
                "RDEPEND": "cups? ( >=net-print/cups-2.0 )",
                "IUSE": "cups",
            },
            "x11-libs/goffice-0.10.55": {
                "EAPI": "8",
                "DEPEND": ">=app-text/libspectre-0.2.6:=",
                "RDEPEND": ">=app-text/libspectre-0.2.6:=",
            },
        }

        installed = {
            "app-text/ghostscript-gpl-10.01.1": {
                "EAPI": "8",
                "DEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "RDEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "IUSE": "dbus gtk",
                "USE": "dbus gtk",
            },
            "app-text/libspectre-0.2.11": {
                "EAPI": "8",
                "DEPEND": ">=app-text/ghostscript-gpl-9.53.0:0/10.01=",
                "RDEPEND": ">=app-text/ghostscript-gpl-9.53.0:0/10.01=",
            },
            "net-dns/avahi-0.8-r7": {
                "EAPI": "8",
                "DEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "RDEPEND": "dbus? ( sys-apps/dbus ) gtk? ( x11-libs/gtk+:3 )",
                "IUSE": "dbus gtk",
                "USE": "dbus gtk",
            },
            "net-print/cups-2.4.6": {
                "EAPI": "8",
                "DEPEND": "zeroconf? ( >=net-dns/avahi-0.6.31-r2[dbus] )",
                "RDEPEND": "zeroconf? ( >=net-dns/avahi-0.6.31-r2[dbus] )",
                "IUSE": "zeroconf",
                "USE": "zeroconf",
            },
            "sys-apps/dbus-1.15.6": {
                "EAPI": "8",
            },
            "x11-libs/gtk+-3.24.38": {
                "EAPI": "8",
                "SLOT": "3",
                "DEPEND": "cups? ( >=net-print/cups-2.0 )",
                "RDEPEND": "cups? ( >=net-print/cups-2.0 )",
                "IUSE": "cups",
                "USE": "cups",
            },
            "x11-libs/goffice-0.10.55": {
                "EAPI": "8",
                "DEPEND": ">=app-text/libspectre-0.2.6:0=",
                "RDEPEND": ">=app-text/libspectre-0.2.6:0=",
            },
        }

        world = [
            "x11-libs/goffice",
        ]

        user_config = {
            "make.conf": ('USE="cups dbus gtk zeroconf"',),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True},
                success=True,
                mergelist=[
                    "app-text/ghostscript-gpl-10.01.2",
                    "app-text/libspectre-0.2.12",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--emptytree": True},
                success=True,
                mergelist=[
                    "sys-apps/dbus-1.15.6",
                    "x11-libs/gtk+-3.24.38",
                    "app-text/ghostscript-gpl-10.01.2",
                    "net-dns/avahi-0.8-r7",
                    "net-print/cups-2.4.6",
                    "app-text/libspectre-0.2.12",
                    "x11-libs/goffice-0.10.55",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            world=world,
            user_config=user_config,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1998-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["dbapi"]

import functools
import re
import warnings
from typing import Any, Dict, List, Optional, Tuple
from collections.abc import Sequence

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.dbapi.dep_expand:dep_expand@_dep_expand",
    "portage.dep:Atom,match_from_list,_match_slot",
    "portage.output:colorize",
    "portage.util:cmp_sort_key,writemsg",
    "portage.versions:catsplit,catpkgsplit,vercmp,_pkg_str",
)

from portage.const import MERGING_IDENTIFIER

from portage import os
from portage import auxdbkeys
from portage.eapi import _get_eapi_attrs
from portage.exception import (
    CorruptionKeyError,
    InvalidBinaryPackageFormat,
    InvalidData,
)
from portage.localization import _
from _emerge.Package import Package


class dbapi:
    _category_re = re.compile(r"^\w[-.+\w]*$", re.UNICODE)
    _categories: Optional[tuple[str, ...]] = None
    _use_mutable = False
    _known_keys = frozenset(auxdbkeys)
    _pkg_str_aux_keys = ("EAPI", "KEYWORDS", "SLOT", "repository")

    def __init__(self):
        pass

    @property
    def categories(self) -> tuple[str, ...]:
        """
        Use self.cp_all() to generate a category list. Mutable instances
        can delete the self._categories attribute in cases when the cached
        categories become invalid and need to be regenerated.
        """
        if self._categories is not None:
            return self._categories
        self._categories = tuple(sorted({catsplit(x)[0] for x in self.cp_all()}))
        return self._categories

    def close_caches(self):
        pass

    def cp_list(self, cp: str, use_cache: int = 1) -> Any:
        raise NotImplementedError(self)

    @staticmethod
    def _cmp_cpv(cpv1, cpv2) -> int:
        result = vercmp(cpv1.version, cpv2.version)
        if result == 0 and cpv1.build_time is not None and cpv2.build_time is not None:
            result = (cpv1.build_time > cpv2.build_time) - (
                cpv1.build_time < cpv2.build_time
            )
        return result

    @staticmethod
    def _cpv_sort_ascending(cpv_list: Sequence[Any]) -> None:
        """
        Use this to sort self.cp_list() results in ascending
        order. It sorts in place and returns None.
        """
        if len(cpv_list) > 1:
            # If the cpv includes explicit -r0, it has to be preserved
            # for consistency in findname and aux_get calls, so use a
            # dict to map strings back to their original values.
            cpv_list.sort(key=cmp_sort_key(dbapi._cmp_cpv))

    def cpv_all(self) -> list[str]:
        """Return all CPVs in the db
        Args:
                None
        Returns:
                A list of Strings, 1 per CPV

        This function relies on a subclass implementing cp_all, this is why the hasattr is there
        """

        if not hasattr(self, "cp_all"):
            raise NotImplementedError
        cpv_list = []
        for cp in self.cp_all():
            cpv_list.extend(self.cp_list(cp))
        return cpv_list

    def cp_all(self, sort: bool = False) -> list[str]:
        """Implement this in a child class
        Args
                sort - return sorted results
        Returns:
                A list of strings 1 per CP in the datastore
        """
        raise NotImplementedError

    def aux_get(
        self, mycpv: str, mylist: str, myrepo: Optional[str] = None
    ) -> list[str]:
        """Return the metadata keys in mylist for mycpv
        Args:
                mycpv - "sys-apps/foo-1.0"
                mylist - ["SLOT","DEPEND","HOMEPAGE"]
                myrepo - The repository name.
        Returns:
                a list of results, in order of keys in mylist, such as:
                ["0",">=sys-libs/bar-1.0","http://www.foo.com"] or [] if mycpv not found'
        """
        raise NotImplementedError

    def aux_update(self, cpv: str, metadata_updates: dict[str, Any]) -> None:
        """
        Args:
          cpv - "sys-apps/foo-1.0"
                metadata_updates = { key : newvalue }
        Returns:
                None
        """
        raise NotImplementedError

    def match(self, origdep: str, use_cache: int = 1):
        """Given a dependency, try to find packages that match
        Args:
                origdep - Depend atom
                use_cache - Boolean indicating if we should use the cache or not
                NOTE: Do we ever not want the cache?
        Returns:
                a list of packages that match origdep
        """
        mydep = _dep_expand(origdep, mydb=self, settings=self.settings)
        return list(
            self._iter_match(mydep, self.cp_list(mydep.cp, use_cache=use_cache))
        )

    def _iter_match(self, atom: str, cpv_iter):
        cpv_iter = iter(match_from_list(atom, cpv_iter))
        if atom.repo:
            cpv_iter = self._iter_match_repo(atom, cpv_iter)
        if atom.slot:
            cpv_iter = self._iter_match_slot(atom, cpv_iter)
        if atom.unevaluated_atom.use:
            cpv_iter = self._iter_match_use(atom, cpv_iter)
        return cpv_iter

    def _pkg_str(self, cpv, repo):
        """
        This is used to construct _pkg_str instances on-demand during
        matching. If cpv is a _pkg_str instance with slot attribute,
        then simply return it. Otherwise, fetch metadata and construct
        a _pkg_str instance. This may raise KeyError or InvalidData.
        """
        try:
            cpv.slot
        except AttributeError:
            pass
        else:
            return cpv

        metadata = dict(
            zip(
                self._pkg_str_aux_keys,
                self.aux_get(cpv, self._pkg_str_aux_keys, myrepo=repo),
            )
        )

        return _pkg_str(cpv, metadata=metadata, settings=self.settings, db=self)

    def _iter_match_repo(self, atom, cpv_iter):
        for cpv in cpv_iter:
            try:
                pkg_str = self._pkg_str(cpv, atom.repo)
            except (KeyError, InvalidData):
                pass
            else:
                if pkg_str.repo == atom.repo:
                    yield pkg_str

    def _iter_match_slot(self, atom, cpv_iter):
        for cpv in cpv_iter:
            try:
                pkg_str = self._pkg_str(cpv, atom.repo)
            except (KeyError, InvalidData):
                pass
            else:
                if _match_slot(atom, pkg_str):
                    yield pkg_str

    def _iter_match_use(self, atom, cpv_iter):
        """
        1) Check for required IUSE intersection (need implicit IUSE here).
        2) Check enabled/disabled flag states.
        """

        aux_keys = ["EAPI", "IUSE", "KEYWORDS", "SLOT", "USE", "repository"]
        for cpv in cpv_iter:
            try:
                metadata = dict(
                    zip(aux_keys, self.aux_get(cpv, aux_keys, myrepo=atom.repo))
                )
            except KeyError:
                continue

            try:
                cpv.slot
            except AttributeError:
                try:
                    cpv = _pkg_str(cpv, metadata=metadata, settings=self.settings)
                except InvalidData:
                    continue

            if not self._match_use(atom, cpv, metadata):
                continue

            yield cpv

    @staticmethod
    def _iuse_implicit_built(iuse_implicit_match, use, flag):
        return iuse_implicit_match(flag) or flag in use

    def _iuse_implicit_cnstr(self, pkg, metadata):
        """
        Construct a callable that checks if a given USE flag should
        be considered to be a member of the implicit IUSE for the
        given package.

        @param pkg: package
        @type pkg: _pkg_str
        @param metadata: package metadata
        @type metadata: Mapping
        @return: a callable that accepts a single USE flag argument,
                and returns True only if the USE flag should be considered
                to be a member of the implicit IUSE for the given package.
        @rtype: callable
        """
        eapi_attrs = _get_eapi_attrs(metadata["EAPI"])
        if eapi_attrs.iuse_effective:
            iuse_implicit_match = self.settings._iuse_effective_match
        else:
            iuse_implicit_match = self.settings._iuse_implicit_match

        if not self._use_mutable and eapi_attrs.iuse_effective:
            # For built packages, it is desirable for the built USE setting to
            # be independent of the profile's current IUSE_IMPLICIT state, since
            # the profile's IUSE_IMPLICT setting may have diverged. Therefore,
            # any member of the built USE setting is considered to be a valid
            # member of IUSE_EFFECTIVE. Note that the binary package may be
            # remote, so it's only possible to rely on metadata that is available
            # in the remote Packages file, and the IUSE_IMPLICIT header in the
            # Packages file is vulnerable to mutation (see bug 640318).
            #
            # This behavior is only used for EAPIs that support IUSE_EFFECTIVE,
            # since built USE settings for earlier EAPIs may contain a large
            # number of irrelevant flags.
            iuse_implicit_match = functools.partial(
                self._iuse_implicit_built,
                iuse_implicit_match,
                frozenset(metadata["USE"].split()),
            )

        return iuse_implicit_match

    def _match_use(self, atom, pkg, metadata, ignore_profile=False):
        iuse_implicit_match = self._iuse_implicit_cnstr(pkg, metadata)
        iuse = Package._iuse(
            None,
            metadata["IUSE"].split(),
            iuse_implicit_match,
            metadata["EAPI"],
        )

        for x in atom.unevaluated_atom.use.required:
            if iuse.get_flag(x) is None:
                return False

        if atom.use is None:
            pass

        elif not self._use_mutable:
            # Use IUSE to validate USE settings for built packages,
            # in case the package manager that built this package
            # failed to do that for some reason (or in case of
            # data corruption). The enabled flags must be consistent
            # with implicit IUSE, in order to avoid potential
            # inconsistencies in USE dep matching (see bug #453400).
            use = frozenset(
                x for x in metadata["USE"].split() if iuse.get_flag(x) is not None
            )
            missing_enabled = frozenset(
                x for x in atom.use.missing_enabled if iuse.get_flag(x) is None
            )
            missing_disabled = frozenset(
                x for x in atom.use.missing_disabled if iuse.get_flag(x) is None
            )
            enabled = frozenset((iuse.get_flag(x) or x) for x in atom.use.enabled)
            disabled = frozenset((iuse.get_flag(x) or x) for x in atom.use.disabled)

            if enabled:
                if any(x in enabled for x in missing_disabled):
                    return False
                need_enabled = enabled.difference(use)
                if need_enabled:
                    if any(x not in missing_enabled for x in need_enabled):
                        return False

            if disabled:
                if any(x in disabled for x in missing_enabled):
                    return False
                need_disabled = disabled.intersection(use)
                if need_disabled:
                    if any(x not in missing_disabled for x in need_disabled):
                        return False

        elif not self.settings.local_config:
            if not ignore_profile:
                # Check masked and forced flags for repoman.
                usemask = self.settings._getUseMask(
                    pkg, stable=self.settings._parent_stable
                )
                if any(
                    x in usemask and iuse.get_flag(x) is not None
                    for x in atom.use.enabled
                ):
                    return False

                useforce = self.settings._getUseForce(
                    pkg, stable=self.settings._parent_stable
                )
                if any(
                    x in useforce and x not in usemask and iuse.get_flag(x) is not None
                    for x in atom.use.disabled
                ):
                    return False

            # Check unsatisfied use-default deps
            if atom.use.enabled:
                missing_disabled = frozenset(
                    x for x in atom.use.missing_disabled if iuse.get_flag(x) is None
                )
                if any(x in atom.use.enabled for x in missing_disabled):
                    return False
            if atom.use.disabled:
                missing_enabled = frozenset(
                    x for x in atom.use.missing_enabled if iuse.get_flag(x) is None
                )
                if any(x in atom.use.disabled for x in missing_enabled):
                    return False

        return True

    def invalidentry(self, mypath):
        if "/" + MERGING_IDENTIFIER in mypath:
            if os.path.exists(mypath):
                writemsg(
                    colorize("BAD", _("INCOMPLETE MERGE:")) + f" {mypath}\n",
                    noiselevel=-1,
                )
        else:
            writemsg(f"!!! Invalid db entry: {mypath}\n", noiselevel=-1)

    def update_ents(self, updates, onProgress=None, onUpdate=None):
        """
        Update metadata of all packages for package moves.
        @param updates: A list of move commands, or dict of {repo_name: list}
        @type updates: list or dict
        @param onProgress: A progress callback function
        @type onProgress: a callable that takes 2 integer arguments: maxval and curval
        @param onUpdate: A progress callback function called only
                for packages that are modified by updates.
        @type onUpdate: a callable that takes 2 integer arguments:
                maxval and curval
        """
        cpv_all = self.cpv_all()
        cpv_all.sort()
        maxval = len(cpv_all)
        aux_get = self.aux_get
        aux_update = self.aux_update
        update_keys = Package._dep_keys
        meta_keys = update_keys + self._pkg_str_aux_keys
        repo_dict = None
        if isinstance(updates, dict):
            repo_dict = updates
        if onUpdate:
            onUpdate(maxval, 0)
        if onProgress:
            onProgress(maxval, 0)
        for i, cpv in enumerate(cpv_all):
            try:
                metadata = dict(zip(meta_keys, aux_get(cpv, meta_keys)))
            except KeyError:
                continue
            try:
                pkg = _pkg_str(cpv, metadata=metadata, settings=self.settings)
            except InvalidData:
                continue
            metadata = {k: metadata[k] for k in update_keys}
            if repo_dict is None:
                updates_list = updates
            else:
                try:
                    updates_list = repo_dict[pkg.repo]
                except KeyError:
                    try:
                        updates_list = repo_dict["DEFAULT"]
                    except KeyError:
                        continue

            if not updates_list:
                continue

            metadata_updates = portage.update_dbentries(
                updates_list, metadata, parent=pkg
            )
            if metadata_updates:
                try:
                    aux_update(cpv, metadata_updates)
                except (InvalidBinaryPackageFormat, CorruptionKeyError) as e:
                    warnings.warn(e)
                if onUpdate:
                    onUpdate(maxval, i + 1)
            if onProgress:
                onProgress(maxval, i + 1)

    def move_slot_ent(self, mylist, repo_match=None):
        """This function takes a sequence:
        Args:
                mylist: a sequence of (atom, originalslot, newslot)
                repo_match: callable that takes single repo_name argument
                        and returns True if the update should be applied
        Returns:
                The number of slotmoves this function did
        """
        atom = mylist[1]
        origslot = mylist[2]
        newslot = mylist[3]

        try:
            atom.with_slot
        except AttributeError:
            atom = Atom(atom).with_slot(origslot)
        else:
            atom = atom.with_slot(origslot)

        origmatches = self.match(atom)
        moves = 0
        if not origmatches:
            return moves
        for mycpv in origmatches:
            try:
                mycpv = self._pkg_str(mycpv, atom.repo)
            except (KeyError, InvalidData):
                continue
            if repo_match is not None and not repo_match(mycpv.repo):
                continue
            moves += 1
            if (
                "/" not in newslot
                and mycpv.sub_slot
                and mycpv.sub_slot not in (mycpv.slot, newslot)
            ):
                newslot = f"{newslot}/{mycpv.sub_slot}"
            mydata = {"SLOT": newslot + "\n"}
            try:
                self.aux_update(mycpv, mydata)
            except CorruptionKeyError as e:
                warnings.warn(e)
                continue
        return moves



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import copy


class BacktrackParameter:
    __slots__ = (
        "circular_dependency",
        "needed_unstable_keywords",
        "runtime_pkg_mask",
        "needed_use_config_changes",
        "needed_license_changes",
        "prune_rebuilds",
        "rebuild_list",
        "reinstall_list",
        "needed_p_mask_changes",
        "slot_operator_mask_built",
        "slot_operator_replace_installed",
    )

    def __init__(self):
        self.circular_dependency = {}
        self.needed_unstable_keywords = set()
        self.needed_p_mask_changes = set()
        self.runtime_pkg_mask = {}
        self.needed_use_config_changes = {}
        self.needed_license_changes = {}
        self.rebuild_list = set()
        self.reinstall_list = set()
        self.slot_operator_replace_installed = set()
        self.slot_operator_mask_built = set()
        self.prune_rebuilds = False

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        result = BacktrackParameter()
        memo[id(self)] = result

        # Shallow copies are enough here, as we only need to ensure that nobody adds stuff
        # to our sets and dicts. The existing content is immutable.
        result.circular_dependency = copy.copy(self.circular_dependency)
        result.needed_unstable_keywords = copy.copy(self.needed_unstable_keywords)
        result.needed_p_mask_changes = copy.copy(self.needed_p_mask_changes)
        result.needed_use_config_changes = copy.copy(self.needed_use_config_changes)
        result.needed_license_changes = copy.copy(self.needed_license_changes)
        result.rebuild_list = copy.copy(self.rebuild_list)
        result.reinstall_list = copy.copy(self.reinstall_list)
        result.slot_operator_replace_installed = copy.copy(
            self.slot_operator_replace_installed
        )
        result.slot_operator_mask_built = self.slot_operator_mask_built.copy()
        result.prune_rebuilds = self.prune_rebuilds

        # runtime_pkg_mask contains nested dicts that must also be copied
        result.runtime_pkg_mask = {}
        for k, v in self.runtime_pkg_mask.items():
            result.runtime_pkg_mask[k] = copy.copy(v)

        return result

    def __eq__(self, other):
        return (
            self.circular_dependency == other.circular_dependency
            and self.needed_unstable_keywords == other.needed_unstable_keywords
            and self.needed_p_mask_changes == other.needed_p_mask_changes
            and self.runtime_pkg_mask == other.runtime_pkg_mask
            and self.needed_use_config_changes == other.needed_use_config_changes
            and self.needed_license_changes == other.needed_license_changes
            and self.rebuild_list == other.rebuild_list
            and self.reinstall_list == other.reinstall_list
            and self.slot_operator_replace_installed
            == other.slot_operator_replace_installed
            and self.slot_operator_mask_built == other.slot_operator_mask_built
            and self.prune_rebuilds == other.prune_rebuilds
        )


class _BacktrackNode:
    __slots__ = (
        "parameter",
        "depth",
        "mask_steps",
        "terminal",
    )

    def __init__(
        self, parameter=BacktrackParameter(), depth=0, mask_steps=0, terminal=True
    ):
        self.parameter = parameter
        self.depth = depth
        self.mask_steps = mask_steps
        self.terminal = terminal

    def __eq__(self, other):
        return self.parameter == other.parameter


class Backtracker:
    __slots__ = (
        "_max_depth",
        "_unexplored_nodes",
        "_current_node",
        "_nodes",
        "_root",
    )

    def __init__(self, max_depth):
        self._max_depth = max_depth
        self._unexplored_nodes = []
        self._current_node = None
        self._nodes = []

        self._root = _BacktrackNode()
        self._add(self._root)

    def _add(self, node, explore=True):
        """
        Adds a newly computed backtrack parameter. Makes sure that it doesn't already exist and
        that we don't backtrack deeper than we are allowed by --backtrack.
        """
        if not self._check_runtime_pkg_mask(node.parameter.runtime_pkg_mask):
            return

        if node.mask_steps <= self._max_depth and node not in self._nodes:
            if explore:
                self._unexplored_nodes.append(node)
            self._nodes.append(node)

    def get(self):
        """
        Returns a backtrack parameter. The backtrack graph is explored with depth first.
        """
        if self._unexplored_nodes:
            node = self._unexplored_nodes.pop()
            self._current_node = node
            return copy.deepcopy(node.parameter)
        return None

    def __len__(self):
        return len(self._unexplored_nodes)

    def _check_runtime_pkg_mask(self, runtime_pkg_mask):
        """
        If a package gets masked that caused other packages to be masked
        before, we revert the mask for other packages (bug 375573).
        """

        for pkg, mask_info in runtime_pkg_mask.items():
            if (
                "missing dependency" in mask_info
                or "slot_operator_mask_built" in mask_info
            ):
                continue

            entry_is_valid = False
            any_conflict_parents = False

            for ppkg, patom in runtime_pkg_mask[pkg].get("slot conflict", set()):
                any_conflict_parents = True
                if ppkg not in runtime_pkg_mask:
                    entry_is_valid = True
                    break
            else:
                if not any_conflict_parents:
                    # Even though pkg was involved in a slot conflict
                    # where it was matched by all involved parent atoms,
                    # consider masking it in order to avoid a missed
                    # update as in bug 692746.
                    entry_is_valid = True

            if not entry_is_valid:
                return False

        return True

    def _feedback_slot_conflicts(self, conflicts_data):
        # Only create BacktrackNode instances for the first
        # conflict which occurred, since the conflicts that
        # occurred later may have been caused by the first
        # conflict.
        self._feedback_slot_conflict(conflicts_data[0])

    def _feedback_slot_conflict(self, conflict_data):
        for similar_pkgs in conflict_data:
            new_node = copy.deepcopy(self._current_node)
            new_node.depth += 1
            new_node.mask_steps += 1
            new_node.terminal = False
            for pkg, parent_atoms in similar_pkgs:
                new_node.parameter.runtime_pkg_mask.setdefault(pkg, {})[
                    "slot conflict"
                ] = parent_atoms
            self._add(new_node)

    def _feedback_missing_dep(self, dep):
        new_node = copy.deepcopy(self._current_node)
        new_node.depth += 1
        new_node.mask_steps += 1
        new_node.terminal = False

        new_node.parameter.runtime_pkg_mask.setdefault(dep.parent, {})[
            "missing dependency"
        ] = {(dep.parent, dep.root, dep.atom)}

        self._add(new_node)

    def _feedback_config(self, changes, explore=True):
        """
        Handle config changes. Don't count config changes for the maximum backtrack depth.
        """
        new_node = copy.deepcopy(self._current_node)
        new_node.depth += 1
        para = new_node.parameter

        for change, data in changes.items():
            if change == "circular_dependency":
                for pkg, circular_children in data.items():
                    para.circular_dependency.setdefault(pkg, set()).update(
                        circular_children
                    )
            elif change == "needed_unstable_keywords":
                para.needed_unstable_keywords.update(data)
            elif change == "needed_p_mask_changes":
                para.needed_p_mask_changes.update(data)
            elif change == "needed_license_changes":
                for pkg, missing_licenses in data:
                    para.needed_license_changes.setdefault(pkg, set()).update(
                        missing_licenses
                    )
            elif change == "needed_use_config_changes":
                for pkg, (new_use, new_changes) in data:
                    para.needed_use_config_changes[pkg] = (new_use, new_changes)
            elif change == "slot_conflict_abi":
                new_node.terminal = False
            elif change == "slot_operator_mask_built":
                para.slot_operator_mask_built.update(data)
                for pkg, mask_reasons in data.items():
                    para.runtime_pkg_mask.setdefault(pkg, {}).update(mask_reasons)
            elif change == "slot_operator_replace_installed":
                para.slot_operator_replace_installed.update(data)
            elif change == "rebuild_list":
                para.rebuild_list.update(data)
            elif change == "reinstall_list":
                para.reinstall_list.update(data)
            elif change == "prune_rebuilds":
                para.prune_rebuilds = True
                para.slot_operator_replace_installed.clear()
                for pkg in para.slot_operator_mask_built:
                    runtime_masks = para.runtime_pkg_mask.get(pkg)
                    if runtime_masks is None:
                        continue
                    runtime_masks.pop("slot_operator_mask_built", None)
                    if not runtime_masks:
                        para.runtime_pkg_mask.pop(pkg)
                para.slot_operator_mask_built.clear()

        self._add(new_node, explore=explore)
        self._current_node = new_node

    def feedback(self, infos):
        """
        Takes information from the depgraph and computes new backtrack parameters to try.
        """
        assert (
            self._current_node is not None
        ), "call feedback() only after get() was called"

        # Not all config changes require a restart, that's why they can appear together
        # with other conflicts.
        if "config" in infos:
            self._feedback_config(infos["config"], explore=len(infos) == 1)

        # There is at most one of the following types of conflicts for a given restart.
        if "slot conflict" in infos:
            self._feedback_slot_conflicts(infos["slot conflict"])
        elif "missing dependency" in infos:
            self._feedback_missing_dep(infos["missing dependency"])

    def backtracked(self):
        """
        If we didn't backtrack, there is only the root.
        """
        return len(self._nodes) > 1

    def get_best_run(self):
        """
        Like, get() but returns the backtrack parameter that has as many config changes as possible,
        but has no masks. This makes --autounmask effective, but prevents confusing error messages
        with "masked by backtracking".
        """
        best_node = self._root
        for node in self._nodes:
            if node.terminal and node.depth > best_node.depth:
                best_node = node

        return copy.deepcopy(best_node.parameter)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.output import bold, turquoise, green


def emerge_help():
    print(bold("emerge:") + " command-line interface to the Portage system")
    print(bold("Usage:"))
    print(
        "   "
        + turquoise("emerge")
        + " [ "
        + green("options")
        + " ] [ "
        + green("action")
        + " ] [ "
        + turquoise("ebuild")
        + " | "
        + turquoise("tbz2")
        + " | "
        + turquoise("file")
        + " | "
        + turquoise("@set")
        + " | "
        + turquoise("atom")
        + " ] [ ... ]"
    )
    print(
        "   "
        + turquoise("emerge")
        + " [ "
        + green("options")
        + " ] [ "
        + green("action")
        + " ] < "
        + turquoise("@system")
        + " | "
        + turquoise("@world")
        + " >"
    )
    print(
        "   "
        + turquoise("emerge")
        + " < "
        + turquoise("--sync")
        + " | "
        + turquoise("--metadata")
        + " | "
        + turquoise("--info")
        + " >"
    )
    print(
        "   "
        + turquoise("emerge")
        + " "
        + turquoise("--resume")
        + " [ "
        + green("--pretend")
        + " | "
        + green("--ask")
        + " | "
        + green("--skipfirst")
        + " ]"
    )
    print("   " + turquoise("emerge") + " " + turquoise("--help"))
    print(
        bold("Options:")
        + " "
        + green("-")
        + "["
        + green("abBcCdDefgGhjkKlnNoOpPqrsStuUvVwW")
        + "]"
    )
    print(
        "          [ "
        + green("--color")
        + " < "
        + turquoise("y")
        + " | "
        + turquoise("n")
        + " >            ] [ "
        + green("--columns")
        + "    ]"
    )
    print(
        "          [ "
        + green("--complete-graph")
        + "             ] [ "
        + green("--deep")
        + "       ]"
    )
    print(
        "          [ "
        + green("--jobs")
        + " "
        + turquoise("JOBS")
        + " ] [ "
        + green("--keep-going")
        + " ] [ "
        + green("--load-average")
        + " "
        + turquoise("LOAD")
        + "            ]"
    )
    print(
        "          [ "
        + green("--newrepo")
        + "   ] [ "
        + green("--newuse")
        + "     ] [ "
        + green("--noconfmem")
        + "  ] [ "
        + green("--nospinner")
        + "   ]"
    )
    print(
        "          [ "
        + green("--oneshot")
        + "   ] [ "
        + green("--onlydeps")
        + "   ] [ "
        + green("--quiet-build")
        + " [ "
        + turquoise("y")
        + " | "
        + turquoise("n")
        + " ]        ]"
    )
    print(
        "          [ "
        + green("--reinstall ")
        + turquoise("changed-use")
        + "      ] [ "
        + green("--with-bdeps")
        + " < "
        + turquoise("y")
        + " | "
        + turquoise("n")
        + " >         ]"
    )
    print(
        bold("Actions:")
        + "  [ "
        + green("--depclean")
        + " | "
        + green("--list-sets")
        + " | "
        + green("--search")
        + " | "
        + green("--sync")
        + " | "
        + green("--version")
        + "        ]"
    )
    print()
    print("   For more help consult the man page.")



# Copyright 2017, 2023 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import shutil
import tempfile

import pytest

from portage import os
from portage.tests import TestCase
from portage.checksum import perform_md5
from portage.util.file_copy import copyfile


class CopyFileTestCase(TestCase):
    def testCopyFile(self):
        tempdir = tempfile.mkdtemp()
        try:
            src_path = os.path.join(tempdir, "src")
            dest_path = os.path.join(tempdir, "dest")
            content = b"foo"

            with open(src_path, "wb") as f:
                f.write(content)

            copyfile(src_path, dest_path)

            self.assertEqual(perform_md5(src_path), perform_md5(dest_path))
        finally:
            shutil.rmtree(tempdir)


class CopyFileSparseTestCase(TestCase):
    def testCopyFileSparse(self):
        tempdir = tempfile.mkdtemp()
        try:
            src_path = os.path.join(tempdir, "src")
            dest_path = os.path.join(tempdir, "dest")
            content = b"foo"

            # Use seek to create some sparse blocks. Don't make these
            # files too big, in case the filesystem doesn't support
            # sparse files.
            with open(src_path, "wb") as f:
                f.write(content)
                f.seek(2**17, 1)
                f.write(content)
                f.seek(2**18, 1)
                f.write(content)
                # Test that sparse blocks are handled correctly at
                # the end of the file (involves seek and truncate).
                f.seek(2**17, 1)

            copyfile(src_path, dest_path)

            self.assertEqual(perform_md5(src_path), perform_md5(dest_path))

            # This last part of the test is expected to fail when sparse
            # copy is not implemented, so mark it xfail:
            pytest.xfail(reason="sparse copy is not implemented")

            # If sparse blocks were preserved, then both files should
            # consume the same number of blocks.
            self.assertEqual(os.stat(src_path).st_blocks, os.stat(dest_path).st_blocks)
        finally:
            shutil.rmtree(tempdir)



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


from portage.dep.soname.SonameAtom import SonameAtom
from portage.tests import TestCase
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor


class TestSonameAtomPickle(TestCase):
    _ALL_PROVIDES = frozenset([SonameAtom("x86_64", "libc.so.6")])

    def test_soname_atom_pickle(self):
        loop = asyncio._wrap_loop()
        with ForkExecutor(loop=loop) as executor:
            result = loop.run_until_complete(
                loop.run_in_executor(executor, self._get_all_provides)
            )
        self.assertEqual(self._ALL_PROVIDES, result)

    @classmethod
    def _get_all_provides(cls):
        return cls._ALL_PROVIDES



# Copyright 2012-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import fcntl
import functools
import multiprocessing
import warnings
import signal
import sys

import portage
from portage import os
from portage.cache.mappings import slot_dict_class
from portage.util.futures import asyncio
from _emerge.SpawnProcess import SpawnProcess


class ForkProcess(SpawnProcess):
    # NOTE: This class overrides the meaning of the SpawnProcess 'args'
    # attribute, and uses it to hold the positional arguments for the
    # 'target' function.
    __slots__ = (
        "kwargs",
        "target",
        "_child_connection",
        "_proc",
        "_proc_join_task",
    )

    _file_names = ("connection", "slave_fd")
    _files_dict = slot_dict_class(_file_names, prefix="")

    # Number of seconds between poll attempts for process exit status
    # (after the sentinel has become ready).
    _proc_join_interval = 0.1

    _HAVE_SEND_HANDLE = getattr(multiprocessing.reduction, "HAVE_SEND_HANDLE", False)

    def _start(self):
        if multiprocessing.get_start_method() == "fork":
            # Backward compatibility mode.
            super()._start()
            return

        # This mode supports multiprocessing start methods
        # other than fork. Note that the fd_pipes implementation
        # uses a thread via run_in_executor, and threads are not
        # recommended for mixing with the fork start method due
        # to cpython issue 84559.
        if self.fd_pipes and not self._HAVE_SEND_HANDLE:
            raise NotImplementedError(
                'fd_pipes only supported with HAVE_SEND_HANDLE or multiprocessing start method "fork"'
            )

        if self.fd_pipes or self.logfile or not self.background:
            # Log via multiprocessing.Pipe if necessary.
            connection, self._child_connection = multiprocessing.Pipe(
                duplex=self._HAVE_SEND_HANDLE
            )

        retval = self._spawn(self.args, fd_pipes=self.fd_pipes)

        self.pid = retval[0]
        self._registered = True

        if self._child_connection is None:
            self._async_waitpid()
        else:
            self._child_connection.close()
            self.fd_pipes = self.fd_pipes or {}
            stdout_fd = None
            if not self.background:
                self.fd_pipes.setdefault(0, portage._get_stdin().fileno())
                self.fd_pipes.setdefault(1, sys.__stdout__.fileno())
                self.fd_pipes.setdefault(2, sys.__stderr__.fileno())
                stdout_fd = os.dup(self.fd_pipes[1])

            if self._HAVE_SEND_HANDLE:
                master_fd, slave_fd = self._pipe(self.fd_pipes)
                self.fd_pipes[1] = slave_fd
                self.fd_pipes[2] = slave_fd
                self._files = self._files_dict(connection=connection, slave_fd=slave_fd)
            else:
                master_fd = connection

            self._start_main_task(
                master_fd, log_file_path=self.logfile, stdout_fd=stdout_fd
            )

    @property
    def _fd_pipes_send_handle(self):
        """Returns True if we have a connection to implement fd_pipes via send_handle."""
        return bool(
            self._HAVE_SEND_HANDLE
            and self._files
            and getattr(self._files, "connection", False)
        )

    def _send_fd_pipes(self):
        """
        Communicate with _bootstrap to send fd_pipes via send_handle.
        This performs blocking IO, intended for invocation via run_in_executor.
        """
        fd_list = list(set(self.fd_pipes.values()))
        self._files.connection.send(
            (self.fd_pipes, fd_list),
        )
        for fd in fd_list:
            multiprocessing.reduction.send_handle(
                self._files.connection,
                fd,
                self.pid,
            )

    async def _main(self, build_logger, pipe_logger, loop=None):
        try:
            if self._fd_pipes_send_handle:
                await self.scheduler.run_in_executor(
                    None,
                    self._send_fd_pipes,
                )
        except asyncio.CancelledError:
            self._main_cancel(build_logger, pipe_logger)
            raise
        finally:
            if self._files:
                if hasattr(self._files, "connection"):
                    self._files.connection.close()
                    del self._files.connection
                if hasattr(self._files, "slave_fd"):
                    os.close(self._files.slave_fd)
                    del self._files.slave_fd

        await super()._main(build_logger, pipe_logger, loop=loop)

    def _spawn(self, args, fd_pipes=None, **kwargs):
        """
        Override SpawnProcess._spawn to fork a subprocess that calls
        self._run(). This uses multiprocessing.Process in order to leverage
        any pre-fork and post-fork interpreter housekeeping that it provides,
        promoting a healthy state for the forked interpreter.
        """

        if self.__class__._run is ForkProcess._run:
            # target replaces the deprecated self._run method
            target = self.target
            args = self.args
            kwargs = self.kwargs
        else:
            # _run implementation triggers backward-compatibility mode
            target = self._run
            args = None
            kwargs = None
            warnings.warn(
                'portage.util._async.ForkProcess.ForkProcess._run is deprecated in favor of the "target" parameter',
                UserWarning,
                stacklevel=2,
            )

        # Since multiprocessing.Process closes sys.__stdin__, create a
        # temporary duplicate of fd_pipes[0] so that sys.__stdin__ can
        # be restored in the subprocess, in case this is needed for
        # things like PROPERTIES=interactive support.
        stdin_dup = None
        try:
            stdin_fd = fd_pipes.get(0) if fd_pipes else None
            if stdin_fd is not None and stdin_fd == portage._get_stdin().fileno():
                stdin_dup = os.dup(stdin_fd)
                fcntl.fcntl(
                    stdin_dup, fcntl.F_SETFD, fcntl.fcntl(stdin_fd, fcntl.F_GETFD)
                )
                fd_pipes[0] = stdin_dup

            if self._fd_pipes_send_handle:
                # Handle fd_pipes in _main instead.
                fd_pipes = None

            self._proc = multiprocessing.Process(
                target=self._bootstrap,
                args=(
                    self._child_connection,
                    self._HAVE_SEND_HANDLE,
                    fd_pipes,
                    target,
                    args,
                    kwargs,
                ),
            )
            self._proc.start()
        finally:
            if stdin_dup is not None:
                os.close(stdin_dup)

        self._proc_join_task = asyncio.ensure_future(
            self._proc_join(self._proc, loop=self.scheduler), loop=self.scheduler
        )
        self._proc_join_task.add_done_callback(
            functools.partial(self._proc_join_done, self._proc)
        )

        return [self._proc.pid]

    def _cancel(self):
        if self._proc is None:
            super()._cancel()
        else:
            self._proc.terminate()

    def _async_wait(self):
        if self._proc_join_task is None:
            super()._async_wait()

    def _async_waitpid(self):
        if self._proc_join_task is None:
            super()._async_waitpid()

    async def _proc_join(self, proc, loop=None):
        sentinel_reader = self.scheduler.create_future()
        self.scheduler.add_reader(
            proc.sentinel,
            lambda: sentinel_reader.done() or sentinel_reader.set_result(None),
        )
        try:
            await sentinel_reader
        finally:
            # If multiprocessing.Process supports the close method, then
            # access to proc.sentinel will raise ValueError if the
            # sentinel has been closed. In this case it's not safe to call
            # remove_reader, since the file descriptor may have been closed
            # and then reallocated to a concurrent coroutine. When the
            # close method is not supported, proc.sentinel remains open
            # until proc's finalizer is called.
            try:
                self.scheduler.remove_reader(proc.sentinel)
            except ValueError:
                pass

        # Now that proc.sentinel is ready, poll until process exit
        # status has become available.
        while True:
            proc.join(0)
            if proc.exitcode is not None:
                break
            await asyncio.sleep(self._proc_join_interval, loop=loop)

    def _proc_join_done(self, proc, future):
        future.cancelled() or future.result()
        self._was_cancelled()
        if self.returncode is None:
            self.returncode = proc.exitcode

        self._proc = None
        if hasattr(proc, "close"):
            proc.close()
        self._proc_join_task = None
        self._async_wait()

    def _unregister(self):
        super()._unregister()
        if self._proc is not None:
            if self._proc.is_alive():
                self._proc.terminate()
            self._proc = None
        if self._proc_join_task is not None:
            self._proc_join_task.cancel()
            self._proc_join_task = None

    @staticmethod
    def _bootstrap(child_connection, have_send_handle, fd_pipes, target, args, kwargs):
        # Use default signal handlers in order to avoid problems
        # killing subprocesses as reported in bug #353239.
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # Unregister SIGCHLD handler and wakeup_fd for the parent
        # process's event loop (bug 655656).
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        try:
            wakeup_fd = signal.set_wakeup_fd(-1)
            if wakeup_fd > 0:
                os.close(wakeup_fd)
        except (ValueError, OSError):
            pass

        portage.locks._close_fds()

        if child_connection is not None:
            if have_send_handle:
                fd_pipes, fd_list = child_connection.recv()
                fd_pipes_map = {}
                for fd in fd_list:
                    fd_pipes_map[fd] = multiprocessing.reduction.recv_handle(
                        child_connection
                    )
                child_connection.close()
                for k, v in list(fd_pipes.items()):
                    fd_pipes[k] = fd_pipes_map[v]

            else:
                fd_pipes = fd_pipes or {}
                fd_pipes[sys.stdout.fileno()] = child_connection.fileno()
                fd_pipes[sys.stderr.fileno()] = child_connection.fileno()
                fd_pipes[child_connection.fileno()] = child_connection.fileno()

        if fd_pipes:
            # We don't exec, so use close_fds=False
            # (see _setup_pipes docstring).
            portage.process._setup_pipes(fd_pipes, close_fds=False)

        # Since multiprocessing.Process closes sys.__stdin__ and
        # makes sys.stdin refer to os.devnull, restore it when
        # appropriate.
        if fd_pipes and 0 in fd_pipes:
            # It's possible that sys.stdin.fileno() is already 0,
            # and in that case the above _setup_pipes call will
            # have already updated its identity via dup2. Otherwise,
            # perform the dup2 call now, and also copy the file
            # descriptor flags.
            if sys.stdin.fileno() != 0:
                os.dup2(0, sys.stdin.fileno())
                fcntl.fcntl(
                    sys.stdin.fileno(), fcntl.F_SETFD, fcntl.fcntl(0, fcntl.F_GETFD)
                )
            sys.__stdin__ = sys.stdin

        sys.exit(target(*(args or []), **(kwargs or {})))

    def _run(self):
        """
        Deprecated and replaced with the "target" constructor parameter.
        """
        raise NotImplementedError(self)



# Copyright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import tempfile
import tarfile

from portage import os, shutil
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.gpkg import gpkg


class test_gpkg_large_size_case(TestCase):
    def test_gpkg_large_size(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="gzip"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)
            # Check if filesystem support sparse file
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.truncate(1048576)

            if os.stat(os.path.join(orig_full_path, "test")).st_blocks != 0:
                self.skipTest("Filesystem does not support sparse file")

            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.truncate(10737418240)

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            check_result = test_gpkg._check_pre_image_files(
                os.path.join(tmpdir, "orig")
            )
            self.assertEqual(check_result, (0, 4, 0, 10737418240, 10737418240))

            test_gpkg.compress(os.path.join(tmpdir, "orig"), {"meta": "test"})

            with open(gpkg_file_loc, "rb") as container:
                # container
                self.assertEqual(
                    test_gpkg._get_tar_format(container), tarfile.GNU_FORMAT
                )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


class SlotObject:
    __slots__ = ("__weakref__",)

    def __init__(self, **kwargs):
        classes = [self.__class__]
        while classes:
            c = classes.pop()
            if c is SlotObject:
                continue
            classes.extend(c.__bases__)
            slots = getattr(c, "__slots__", None)
            if not slots:
                continue
            for myattr in slots:
                myvalue = kwargs.pop(myattr, None)
                if myvalue is None and getattr(self, myattr, None) is not None:
                    raise AssertionError(
                        "class '%s' duplicates '%s' value in __slots__ of base class '%s'"
                        % (self.__class__.__name__, myattr, c.__name__)
                    )
                try:
                    setattr(self, myattr, myvalue)
                except AttributeError:
                    # Allow a property to override a __slots__ value, but raise an
                    # error if the intended value is something other than None.
                    if not (
                        myvalue is None
                        and isinstance(getattr(type(self), myattr, None), property)
                    ):
                        raise

        if kwargs:
            raise TypeError(
                "'%s' is an invalid keyword argument for this constructor"
                % (next(iter(kwargs)),)
            )

    def copy(self):
        """
        Create a new instance and copy all attributes
        defined from __slots__ (including those from
        inherited classes).
        """
        obj = self.__class__()

        classes = [self.__class__]
        while classes:
            c = classes.pop()
            if c is SlotObject:
                continue
            classes.extend(c.__bases__)
            slots = getattr(c, "__slots__", None)
            if not slots:
                continue
            for myattr in slots:
                setattr(obj, myattr, getattr(self, myattr))

        return obj



# Copyright 2010-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging

from _emerge.DepPrioritySatisfiedRange import DepPrioritySatisfiedRange
from _emerge.Package import Package

from itertools import chain, product

from portage.dep import (
    use_reduce,
    extract_affecting_use,
    check_required_use,
    get_required_use_flags,
)
from portage.exception import InvalidDependString
from portage.output import colorize
from portage.util import writemsg_level


class circular_dependency_handler:
    MAX_AFFECTING_USE = 10

    def __init__(self, depgraph, graph):
        self.depgraph = depgraph
        self.graph = graph
        self.all_parent_atoms = depgraph._dynamic_config._parent_atoms

        if "--debug" in depgraph._frozen_config.myopts:
            # Show this debug output before doing the calculations
            # that follow, so at least we have this debug info
            # if we happen to hit a bug later.
            writemsg_level(
                "\n\ncircular dependency graph:\n\n", level=logging.DEBUG, noiselevel=-1
            )
            self.debug_print()

        self.cycles, self.shortest_cycle = self._find_cycles()
        # Guess if it is a large cluster of cycles. This usually requires
        # a global USE change.
        self.large_cycle_count = len(self.cycles) > 3
        self.merge_list = self._prepare_reduced_merge_list()
        # The digraph dump
        self.circular_dep_message = self._prepare_circular_dep_message()
        # Suggestions, in machine and human readable form
        self.solutions, self.suggestions = self._find_suggestions()

    def _find_cycles(self):
        shortest_cycle = None
        cycles = self.graph.get_cycles(
            ignore_priority=DepPrioritySatisfiedRange.ignore_medium_soft
        )
        for cycle in cycles:
            if not shortest_cycle or len(cycle) < len(shortest_cycle):
                shortest_cycle = cycle
        return cycles, shortest_cycle

    def _prepare_reduced_merge_list(self):
        """
        Create a merge to be displayed by depgraph.display().
        This merge list contains only packages involved in
        the circular deps.
        """
        display_order = []
        tempgraph = self.graph.copy()
        while tempgraph:
            nodes = tempgraph.leaf_nodes()
            if not nodes:
                node = tempgraph.order[0]
            else:
                node = nodes[0]
            display_order.append(node)
            tempgraph.remove(node)
        return tuple(display_order)

    def _prepare_circular_dep_message(self):
        """
        Like digraph.debug_print(), but prints only the shortest cycle.
        """
        if not self.shortest_cycle:
            return None

        msg = []
        indent = ""
        for pos, pkg in enumerate(self.shortest_cycle):
            parent = self.shortest_cycle[pos - 1]
            priorities = self.graph.nodes[parent][0][pkg]
            if pos > 0:
                msg.append(indent + f"{pkg} ({priorities[-1]})")
            else:
                msg.append(indent + f"{pkg} depends on")
            indent += " "

        pkg = self.shortest_cycle[0]
        parent = self.shortest_cycle[-1]
        priorities = self.graph.nodes[parent][0][pkg]
        msg.append(indent + f"{pkg} ({priorities[-1]})")

        return "\n".join(msg)

    def _get_use_mask_and_force(self, pkg):
        return pkg.use.mask, pkg.use.force

    def _get_autounmask_changes(self, pkg):
        needed_use_config_change = (
            self.depgraph._dynamic_config._needed_use_config_changes.get(pkg)
        )
        if needed_use_config_change is None:
            return frozenset()

        use, changes = needed_use_config_change
        return frozenset(changes.keys())

    def _find_suggestions(self):
        if not self.shortest_cycle:
            return None, None

        suggestions = []
        final_solutions = {}

        for pos, pkg in enumerate(self.shortest_cycle):
            parent = self.shortest_cycle[pos - 1]
            priorities = self.graph.nodes[parent][0][pkg]
            parent_atoms = self.all_parent_atoms.get(pkg)

            if priorities[-1].buildtime:
                dep = " ".join(parent._metadata[k] for k in Package._buildtime_keys)
            elif priorities[-1].runtime:
                dep = parent._metadata["RDEPEND"]

            for ppkg, atom in parent_atoms:
                if ppkg == parent:
                    changed_parent = ppkg
                    parent_atom = atom
                    break

            if parent_atom.package:
                parent_atom = parent_atom.unevaluated_atom
            else:
                # Treat soname deps as unconditional for now. In some
                # cases they can be avoided by a rebuild with changed
                # USE, but ebuilds sometimes do not specify the
                # corresponding conditional dependency (especially for
                # system packages like gcc which provides libstdc++.so.6
                # and libgcc_s.so.1).
                continue

            try:
                affecting_use = extract_affecting_use(
                    dep, parent_atom, eapi=parent.eapi
                )
            except InvalidDependString:
                if not parent.installed:
                    raise
                affecting_use = set()

            # Make sure we don't want to change a flag that is
            # 	a) in use.mask or use.force
            # 	b) changed by autounmask

            usemask, useforce = self._get_use_mask_and_force(parent)
            autounmask_changes = self._get_autounmask_changes(parent)
            untouchable_flags = frozenset(chain(usemask, useforce, autounmask_changes))

            affecting_use.difference_update(untouchable_flags)

            # If any of the flags we're going to touch is in REQUIRED_USE, add all
            # other flags in REQUIRED_USE to affecting_use, to not lose any solution.
            required_use_flags = get_required_use_flags(
                parent._metadata.get("REQUIRED_USE", ""), eapi=parent.eapi
            )

            if affecting_use.intersection(required_use_flags):
                # TODO: Find out exactly which REQUIRED_USE flags are
                # entangled with affecting_use. We have to limit the
                # number of flags since the number of loops is
                # exponentially related (see bug #374397).
                total_flags = set()
                total_flags.update(affecting_use, required_use_flags)
                total_flags.difference_update(untouchable_flags)
                if len(total_flags) <= self.MAX_AFFECTING_USE:
                    affecting_use = total_flags

            affecting_use = tuple(affecting_use)

            if not affecting_use:
                continue

            if len(affecting_use) > self.MAX_AFFECTING_USE:
                # Limit the number of combinations explored (bug #555698).
                # First, discard irrelevent flags that are not enabled.
                # Since extract_affecting_use doesn't distinguish between
                # positive and negative effects (flag? vs. !flag?), assume
                # a positive relationship.
                current_use = self.depgraph._pkg_use_enabled(parent)
                affecting_use = tuple(
                    flag for flag in affecting_use if flag in current_use
                )

                if len(affecting_use) > self.MAX_AFFECTING_USE:
                    # There are too many USE combinations to explore in
                    # a reasonable amount of time.
                    continue

            # We iterate over all possible settings of these use flags and gather
            # a set of possible changes
            # TODO: Use the information encoded in REQUIRED_USE
            solutions = set()
            for use_state in product(
                ("disabled", "enabled"), repeat=len(affecting_use)
            ):
                current_use = set(self.depgraph._pkg_use_enabled(parent))
                for flag, state in zip(affecting_use, use_state):
                    if state == "enabled":
                        current_use.add(flag)
                    else:
                        current_use.discard(flag)
                try:
                    reduced_dep = use_reduce(dep, uselist=current_use, flat=True)
                except InvalidDependString:
                    if not parent.installed:
                        raise
                    reduced_dep = None

                if reduced_dep is not None and parent_atom not in reduced_dep:
                    # We found an assignment that removes the atom from 'dep'.
                    # Make sure it doesn't conflict with REQUIRED_USE.
                    required_use = parent._metadata.get("REQUIRED_USE", "")

                    if check_required_use(
                        required_use,
                        current_use,
                        parent.iuse.is_valid_flag,
                        eapi=parent.eapi,
                    ):
                        use = self.depgraph._pkg_use_enabled(parent)
                        solution = set()
                        for flag, state in zip(affecting_use, use_state):
                            if state == "enabled" and flag not in use:
                                solution.add((flag, True))
                            elif state == "disabled" and flag in use:
                                solution.add((flag, False))
                        solutions.add(frozenset(solution))

            for solution in solutions:
                ignore_solution = False
                for other_solution in solutions:
                    if solution is other_solution:
                        continue
                    if solution.issuperset(other_solution):
                        ignore_solution = True
                if ignore_solution:
                    continue

                # Check if a USE change conflicts with use requirements of the parents.
                # If a requiremnet is hard, ignore the suggestion.
                # If the requirment is conditional, warn the user that other changes might be needed.
                followup_change = False
                parent_parent_atoms = self.depgraph._dynamic_config._parent_atoms.get(
                    changed_parent
                )
                for ppkg, atom in parent_parent_atoms:
                    atom = atom.unevaluated_atom
                    if not atom.use:
                        continue

                    for flag, state in solution:
                        if flag in atom.use.enabled or flag in atom.use.disabled:
                            ignore_solution = True
                            break
                        elif atom.use.conditional:
                            for flags in atom.use.conditional.values():
                                if flag in flags:
                                    followup_change = True
                                    break

                    if ignore_solution:
                        break

                if ignore_solution:
                    continue

                changes = []
                for flag, state in solution:
                    if state:
                        changes.append(colorize("red", "+" + flag))
                    else:
                        changes.append(colorize("blue", "-" + flag))
                msg = f"- {parent.cpv} (Change USE: {' '.join(changes)})\n"
                if followup_change:
                    msg += (
                        " (This change might require USE changes on parent packages.)"
                    )
                suggestions.append(msg)
                final_solutions.setdefault(pkg, set()).add(solution)

        return final_solutions, suggestions

    def debug_print(self):
        """
        Create a copy of the digraph, prune all root nodes,
        and call the debug_print() method.
        """
        graph = self.graph.copy()
        while True:
            root_nodes = graph.root_nodes(
                ignore_priority=DepPrioritySatisfiedRange.ignore_medium_soft
            )
            if not root_nodes:
                break
            graph.difference_update(root_nodes)

        graph.debug_print()



# test_lafilefixer.py -- Portage Unit Testing Functionality
# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.exception import InvalidData


class test_lafilefixer(TestCase):
    def get_test_cases_clean(self):
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -lm'\n" + b"current=6\n" + b"age=0\n" + b"revision=2\n" + b"installed=yes\n" + b"dlopen=''\n" + b"dlpreopen=''\n" + b"libdir='/usr/lib64'\n"
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -lm'\n" + b"current=6\n" + b"age=0\n" + b"revision=2\n" + b"installed=yes\n" + b"dlopen=''\n" + b"dlpreopen=''\n" + b"libdir='/usr/lib64'\n"
        yield b"dependency_libs=' liba.la /usr/lib64/bar.la -lc'\n"

    def get_test_cases_update(self):
        # .la -> -l*
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc'\n", b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -L/usr/lib64 -la -lb -lc'\n"
        # move stuff into inherited_linker_flags
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la -pthread /usr/lib64/libb.la -lc'\n" + b"inherited_linker_flags=''\n", b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -L/usr/lib64 -la -lb -lc'\n" + b"inherited_linker_flags=' -pthread'\n"
        # reorder
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la -R/usr/lib64 /usr/lib64/libb.la -lc'\n", b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -R/usr/lib64 -L/usr/lib64 -la -lb -lc'\n"
        # remove duplicates from dependency_libs (the original version didn't do it for inherited_linker_flags)
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libc.la -pthread -mt" + b" -L/usr/lib -R/usr/lib64 -lc /usr/lib64/libb.la -lc'\n" + b"inherited_linker_flags=' -pthread -pthread'\n", b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' -R/usr/lib64 -L/usr/lib64 -L/usr/lib -la -lc -lb'\n" + b"inherited_linker_flags=' -pthread -pthread -mt'\n"
        # -L rewriting
        yield b"dependency_libs=' -L/usr/X11R6/lib'\n", b"dependency_libs=' -L/usr/lib'\n"
        yield b"dependency_libs=' -L/usr/local/lib'\n", b"dependency_libs=' -L/usr/lib'\n"
        yield b"dependency_libs=' -L/usr/lib64/pkgconfig/../..'\n", b"dependency_libs=' -L/usr'\n"
        yield b"dependency_libs=' -L/usr/lib/pkgconfig/..'\n", b"dependency_libs=' -L/usr/lib'\n"
        yield b"dependency_libs=' -L/usr/lib/pkgconfig/../.. -L/usr/lib/pkgconfig/..'\n", b"dependency_libs=' -L/usr -L/usr/lib'\n"
        # we once got a backtrace on this one
        yield b"dependency_libs=' /usr/lib64/libMagickCore.la -L/usr/lib64 -llcms2 /usr/lib64/libtiff.la " + b"-ljbig -lc /usr/lib64/libfreetype.la /usr/lib64/libjpeg.la /usr/lib64/libXext.la " + b"/usr/lib64/libXt.la /usr/lib64/libSM.la -lICE -luuid /usr/lib64/libICE.la /usr/lib64/libX11.la " + b"/usr/lib64/libxcb.la /usr/lib64/libXau.la /usr/lib64/libXdmcp.la -lbz2 -lz -lm " + b"/usr/lib/gcc/x86_64-pc-linux-gnu/4.4.4/libgomp.la -lrt -lpthread /usr/lib64/libltdl.la -ldl " + b"/usr/lib64/libfpx.la -lstdc++'", b"dependency_libs=' -L/usr/lib64 -L/usr/lib/gcc/x86_64-pc-linux-gnu/4.4.4 -lMagickCore -llcms2 " + b"-ltiff -ljbig -lc -lfreetype -ljpeg -lXext -lXt -lSM -lICE -luuid -lX11 -lxcb -lXau -lXdmcp " + b"-lbz2 -lz -lm -lgomp -lrt -lpthread -lltdl -ldl -lfpx -lstdc++'"

    def get_test_cases_broken(self):
        yield b""
        # no dependency_libs
        yield b"dlname='libfoo.so.1'\n" + b"current=6\n" + b"age=0\n" + b"revision=2\n"
        # broken dependency_libs
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc' \n"
        # broken dependency_libs
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc\n"
        # crap in dependency_libs
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc /-lstdc++'\n"
        # dependency_libs twice
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc /-lstdc++'\n" + b"dependency_libs=' /usr/lib64/liba.la /usr/lib64/libb.la -lc /-lstdc++'\n"
        # inherited_linker_flags twice
        yield b"dlname='libfoo.so.1'\n" + b"library_names='libfoo.so.1.0.2 libfoo.so.1 libfoo.so'\n" + b"old_library='libpdf.a'\n" + b"inherited_linker_flags=''\n" + b"inherited_linker_flags=''\n"

    def testlafilefixer(self):
        from portage.util.lafilefixer import _parse_lafile_contents, rewrite_lafile

        for clean_contents in self.get_test_cases_clean():
            self.assertEqual(rewrite_lafile(clean_contents), (False, None))

        for original_contents, fixed_contents in self.get_test_cases_update():
            self.assertEqual(rewrite_lafile(original_contents), (True, fixed_contents))

        for broken_contents in self.get_test_cases_broken():
            self.assertRaises(InvalidData, rewrite_lafile, broken_contents)



# Copyright 2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from unittest.mock import MagicMock, patch

from _emerge.actions import get_libc_version, run_action

from portage.const import LIBC_PACKAGE_ATOM
from portage.dbapi.virtual import fakedbapi
from portage.dep import Atom
from portage.tests import TestCase


class RunActionTestCase(TestCase):
    """This class' purpose is to encompass UTs for ``actions.run_action``.
    Since that function is extremely long (at least on Sep. 2022;
    hopefully the situation gets better with the time), the tests in this
    ``TestCase`` contain plenty of mocks/patches.
    Hopefully, with time and effort, the ``run_action`` function (and others
    in the module) are refactored to make testing easier and more robust.

    A side effect of the mocking approach is a strong dependency on the
    details of the implementation. That can be improved if functions
    are smaller and do a well defined small set of tasks. Another call to
    refactoring...
    If the implementation changes, the mocks can be adjusted to play its
    role.
    """

    @patch("_emerge.actions.profile_check")
    @patch("_emerge.actions.adjust_configs")
    @patch("_emerge.actions.apply_priorities")
    def test_binary_trees_populate_called(self, papply, padjust, profile_ckeck):
        """Ensure that ``binarytree.populate`` API is correctly used.
        The point of this test is to ensure that the ``populate`` method
        is called as expected: since it is the first time that ``populate``
        is called, it must use ``getbinpkg_refresh=True``.
        """
        config = MagicMock()
        config.action = None
        config.opts = {"--quiet": True, "--usepkg": True, "--package-moves": "n"}
        bt = MagicMock()
        tree = {"bintree": bt}
        trees = {"first": tree}
        config.trees = trees

        run_action(config)

        bt.populate.assert_called_once_with(
            getbinpkgs=False, getbinpkg_refresh=True, pretend=False
        )

    def testGetSystemLibc(self):
        """
        Check that get_libc_version extracts the right version string
        from the provider LIBC_PACKAGE_ATOM for emerge --info and friends.
        """
        settings = MagicMock()

        settings.getvirtuals.return_value = {
            LIBC_PACKAGE_ATOM: [Atom("=sys-libs/musl-1.2.3")]
        }
        settings.__getitem__.return_value = {}

        vardb = fakedbapi(settings)
        vardb.cpv_inject("sys-libs/musl-1.2.3", {"SLOT": "0"})

        self.assertEqual(get_libc_version(vardb), ["musl-1.2.3"])



# Copyright 2004-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""
Portage: Lock management code
"""

__all__ = [
    "lockdir",
    "unlockdir",
    "lockfile",
    "unlockfile",
    "hardlock_name",
    "hardlink_is_mine",
    "hardlink_lockfile",
    "unhardlink_lockfile",
    "hardlock_cleanup",
]

import errno
import fcntl
import functools
import multiprocessing
import sys
import tempfile
import time
import typing
import warnings

import portage
from portage import os, _encodings, _unicode_decode
from portage.exception import (
    DirectoryNotFound,
    FileNotFound,
    InvalidData,
    TryAgain,
    OperationNotPermitted,
    PermissionDenied,
    ReadOnlyFileSystem,
)
from portage.util import writemsg
from portage.util.install_mask import _raise_exc
from portage.localization import _


HARDLINK_FD = -2
_HARDLINK_POLL_LATENCY = 3  # seconds

# Used by emerge in order to disable the "waiting for lock" message
# so that it doesn't interfere with the status display.
_quiet = False


_lock_fn = None
_open_fds = {}
_open_inodes = {}


class _lock_manager:
    __slots__ = ("fd", "inode_key")

    def __init__(self, fd, fstat_result, path):
        self.fd = fd
        self.inode_key = (fstat_result.st_dev, fstat_result.st_ino)
        if self.inode_key in _open_inodes:
            # This means that the lock is already held by the current
            # process, so the caller will have to try again. This case
            # is encountered with the default fcntl.lockf function, and
            # with the alternative fcntl.flock function TryAgain is
            # raised earlier.
            os.close(fd)
            raise TryAgain(path)
        _open_fds[fd] = self
        _open_inodes[self.inode_key] = self

    def close(self):
        os.close(self.fd)
        del _open_fds[self.fd]
        del _open_inodes[self.inode_key]


def _lockf_test_lock_fn(path, fd, flags):
    fcntl.lockf(fd, flags)
    return functools.partial(unlockfile, (path, fd, flags, fcntl.lockf))


def _get_lock_fn():
    """
    Returns fcntl.lockf if proven to work, and otherwise returns fcntl.flock.
    On some platforms fcntl.lockf is known to be broken.
    """
    global _lock_fn
    if _lock_fn is not None:
        return _lock_fn

    if _test_lock_fn(_lockf_test_lock_fn):
        _lock_fn = fcntl.lockf
        return _lock_fn

    # Fall back to fcntl.flock.
    _lock_fn = fcntl.flock
    return _lock_fn


def _test_lock_fn(
    lock_fn: typing.Callable[[str, int, int], typing.Callable[[], None]]
) -> bool:
    fd, lock_path = tempfile.mkstemp()
    unlock_fn = None
    try:
        try:
            unlock_fn = lock_fn(lock_path, fd, fcntl.LOCK_EX)
        except (TryAgain, OSError):
            pass
        else:
            _lock_manager(fd, os.fstat(fd), lock_path)
            proc = multiprocessing.Process(
                target=_subprocess_test_lock,
                args=(
                    # Since file descriptors are not inherited unless the fork start
                    # method is used, the subprocess should only try to close an
                    # inherited file descriptor for the fork start method.
                    fd if multiprocessing.get_start_method() == "fork" else None,
                    lock_fn,
                    lock_path,
                ),
            )
            proc.start()
            proc.join()
            if proc.exitcode == os.EX_OK:
                # the test passed
                return True
    finally:
        try:
            os.unlink(lock_path)
        except OSError:
            pass
        if unlock_fn is not None:
            unlock_fn()
    return False


def _subprocess_test_lock(fd, lock_fn, lock_path):
    if fd is not None:
        os.close(fd)
    try:
        with open(lock_path, "a") as f:
            lock_fn(lock_path, f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
    except (TryAgain, OSError) as e:
        if isinstance(e, TryAgain) or e.errno == errno.EAGAIN:
            # Parent process holds lock, as expected.
            sys.exit(0)

    # Something went wrong.
    sys.exit(1)


def _close_fds():
    """
    This is intended to be called after a fork, in order to close file
    descriptors for locks held by the parent process. This can be called
    safely after a fork without exec, unlike the _setup_pipes close_fds
    behavior.
    """
    for fd in list(_open_fds.values()):
        fd.close()


def lockdir(mydir, flags=0):
    return lockfile(mydir, wantnewlockfile=1, flags=flags)


def unlockdir(mylock):
    return unlockfile(mylock)


def lockfile(mypath, wantnewlockfile=0, unlinkfile=0, waiting_msg=None, flags=0):
    """
    If wantnewlockfile is True then this creates a lockfile in the parent
    directory as the file: '.' + basename + '.portage_lockfile'.
    """
    lock = None
    while lock is None:
        lock = _lockfile_iteration(
            mypath,
            wantnewlockfile=wantnewlockfile,
            unlinkfile=unlinkfile,
            waiting_msg=waiting_msg,
            flags=flags,
        )
        if lock is None:
            writemsg(_("lockfile removed by previous lock holder, retrying\n"), 1)
    return lock


def _lockfile_iteration(
    mypath, wantnewlockfile=False, unlinkfile=False, waiting_msg=None, flags=0
):
    """
    Acquire a lock on mypath, without retry. Return None if the lockfile
    was removed by previous lock holder (caller must retry).

    @param mypath: lock file path
    @type mypath: str
    @param wantnewlockfile: use a separate new lock file
    @type wantnewlockfile: bool
    @param unlinkfile: remove lock file prior to unlock
    @type unlinkfile: bool
    @param waiting_msg: message to show before blocking
    @type waiting_msg: str
    @param flags: lock flags (only supports os.O_NONBLOCK)
    @type flags: int
    @rtype: bool
    @return: unlockfile tuple on success, None if retry is needed
    """
    if not mypath:
        raise InvalidData(_("Empty path given"))

    # Since Python 3.4, chown requires int type (no proxies).
    portage_gid = int(portage.data.portage_gid)

    # Support for file object or integer file descriptor parameters is
    # deprecated due to ambiguity in whether or not it's safe to close
    # the file descriptor, making it prone to "Bad file descriptor" errors
    # or file descriptor leaks.
    if isinstance(mypath, str) and mypath[-1] == "/":
        mypath = mypath[:-1]

    lockfilename_path = mypath
    if hasattr(mypath, "fileno"):
        warnings.warn(
            "portage.locks.lockfile() support for "
            "file object parameters is deprecated. Use a file path instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        lockfilename_path = getattr(mypath, "name", None)
        mypath = mypath.fileno()
    if isinstance(mypath, int):
        warnings.warn(
            "portage.locks.lockfile() support for integer file "
            "descriptor parameters is deprecated. Use a file path instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        lockfilename = mypath
        wantnewlockfile = 0
        unlinkfile = 0
    elif wantnewlockfile:
        base, tail = os.path.split(mypath)
        lockfilename = os.path.join(base, "." + tail + ".portage_lockfile")
        lockfilename_path = lockfilename
        unlinkfile = 1
    else:
        lockfilename = mypath

    if isinstance(mypath, str):
        if not os.path.exists(os.path.dirname(mypath)):
            raise DirectoryNotFound(os.path.dirname(mypath))
        preexisting = os.path.exists(lockfilename)
        old_mask = os.umask(000)
        try:
            while True:
                try:
                    myfd = os.open(lockfilename, os.O_CREAT | os.O_RDWR, 0o660)
                except OSError as e:
                    if e.errno in (errno.ENOENT, errno.ESTALE) and os.path.isdir(
                        os.path.dirname(lockfilename)
                    ):
                        # Retry required for NFS (see bug 636798).
                        continue
                    else:
                        _raise_exc(e)
                else:
                    break

            if not preexisting:
                try:
                    if (
                        portage.data.secpass >= 1
                        and os.stat(lockfilename).st_gid != portage_gid
                    ):
                        os.chown(lockfilename, -1, portage_gid)
                except OSError as e:
                    if e.errno in (errno.ENOENT, errno.ESTALE):
                        os.close(myfd)
                        return None
                    writemsg(
                        "%s: chown('%s', -1, %d)\n" % (e, lockfilename, portage_gid),
                        noiselevel=-1,
                    )
                    writemsg(
                        _("Cannot chown a lockfile: '%s'\n") % lockfilename,
                        noiselevel=-1,
                    )
                    writemsg(
                        _("Group IDs of current user: %s\n")
                        % " ".join(str(n) for n in os.getgroups()),
                        noiselevel=-1,
                    )
        finally:
            os.umask(old_mask)

    elif isinstance(mypath, int):
        myfd = mypath

    else:
        raise ValueError(
            _("Unknown type passed in '%s': '%s'") % (type(mypath), mypath)
        )

    # try for a non-blocking lock, if it's held, throw a message
    # we're waiting on lockfile and use a blocking attempt.
    locking_method = _get_lock_fn()
    try:
        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            raise OSError(errno.ENOSYS, "Function not implemented")
        locking_method(myfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except OSError as e:
        if not hasattr(e, "errno"):
            raise
        if e.errno in (errno.EACCES, errno.EAGAIN, errno.ENOLCK):
            # resource temp unavailable; eg, someone beat us to the lock.
            if flags & os.O_NONBLOCK:
                os.close(myfd)
                raise TryAgain(mypath)

            global _quiet
            if _quiet:
                out = None
            else:
                out = portage.output.EOutput()
            if waiting_msg is None:
                if isinstance(mypath, int):
                    waiting_msg = _("waiting for lock on fd %i") % myfd
                else:
                    waiting_msg = _("waiting for lock on %s") % lockfilename
            if out is not None:
                out.ebegin(waiting_msg)
            # try for the exclusive lock now.
            enolock_msg_shown = False
            while True:
                try:
                    locking_method(myfd, fcntl.LOCK_EX)
                except OSError as e:
                    if e.errno == errno.ENOLCK:
                        # This is known to occur on Solaris NFS (see
                        # bug #462694). Assume that the error is due
                        # to temporary exhaustion of record locks,
                        # and loop until one becomes available.
                        if not enolock_msg_shown:
                            enolock_msg_shown = True
                            if isinstance(mypath, int):
                                context_desc = (
                                    _("Error while waiting " "to lock fd %i") % myfd
                                )
                            else:
                                context_desc = (
                                    _("Error while waiting " "to lock '%s'")
                                    % lockfilename
                                )
                            writemsg(f"\n!!! {context_desc}: {e}\n", noiselevel=-1)

                        time.sleep(_HARDLINK_POLL_LATENCY)
                        continue

                    if out is not None:
                        out.eend(1, str(e))
                    raise
                else:
                    break

            if out is not None:
                out.eend(os.EX_OK)
        elif e.errno in (errno.ENOSYS,):
            # We're not allowed to lock on this FS.
            if not isinstance(lockfilename, int):
                # If a file object was passed in, it's not safe
                # to close the file descriptor because it may
                # still be in use.
                os.close(myfd)
            lockfilename_path = _unicode_decode(
                lockfilename_path, encoding=_encodings["fs"], errors="strict"
            )
            if not isinstance(lockfilename_path, str):
                raise
            link_success = hardlink_lockfile(
                lockfilename_path, waiting_msg=waiting_msg, flags=flags
            )
            if not link_success:
                raise
            lockfilename = lockfilename_path
            locking_method = None
            myfd = HARDLINK_FD
        else:
            raise

    fstat_result = None
    if isinstance(lockfilename, str) and myfd != HARDLINK_FD and unlinkfile:
        try:
            (removed, fstat_result) = _lockfile_was_removed(myfd, lockfilename)
        except Exception:
            # Do not leak the file descriptor here.
            os.close(myfd)
            raise
        else:
            if removed:
                # Removed by previous lock holder... Caller will retry...
                os.close(myfd)
                return None

    if myfd != HARDLINK_FD:
        _lock_manager(
            myfd, os.fstat(myfd) if fstat_result is None else fstat_result, mypath
        )

    writemsg(str((lockfilename, myfd, unlinkfile)) + "\n", 1)
    return (lockfilename, myfd, unlinkfile, locking_method)


def _lockfile_was_removed(lock_fd, lock_path):
    """
    Check if lock_fd still refers to a file located at lock_path, since
    the file may have been removed by a concurrent process that held the
    lock earlier. This implementation includes support for NFS, where
    stat is not reliable for removed files due to the default file
    attribute cache behavior ('ac' mount option).

    @param lock_fd: an open file descriptor for a lock file
    @type lock_fd: int
    @param lock_path: path of lock file
    @type lock_path: str
    @rtype: bool
    @return: a tuple of (removed, fstat_result), where removed is True if
            lock_path does not correspond to lock_fd, and False otherwise
    """
    try:
        fstat_st = os.fstat(lock_fd)
    except OSError as e:
        if e.errno not in (errno.ENOENT, errno.ESTALE):
            _raise_exc(e)
        return (True, None)

    # Since stat is not reliable for removed files on NFS with the default
    # file attribute cache behavior ('ac' mount option), create a temporary
    # hardlink in order to prove that the file path exists on the NFS server.
    hardlink_path = hardlock_name(lock_path)
    try:
        os.unlink(hardlink_path)
    except OSError as e:
        if e.errno not in (errno.ENOENT, errno.ESTALE):
            _raise_exc(e)
    try:
        try:
            os.link(lock_path, hardlink_path)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                _raise_exc(e)
            return (True, None)

        hardlink_stat = os.stat(hardlink_path)
        if (
            hardlink_stat.st_ino != fstat_st.st_ino
            or hardlink_stat.st_dev != fstat_st.st_dev
        ):
            # Create another hardlink in order to detect whether or not
            # hardlink inode numbers are expected to match. For example,
            # inode numbers are not expected to match for sshfs.
            inode_test = hardlink_path + "-inode-test"
            try:
                os.unlink(inode_test)
            except OSError as e:
                if e.errno not in (errno.ENOENT, errno.ESTALE):
                    _raise_exc(e)
            try:
                os.link(hardlink_path, inode_test)
            except OSError as e:
                if e.errno not in (errno.ENOENT, errno.ESTALE):
                    _raise_exc(e)
                return (True, None)
            else:
                if not os.path.samefile(hardlink_path, inode_test):
                    # This implies that inode numbers are not expected
                    # to match for this file system, so use a simple
                    # stat call to detect if lock_path has been removed.
                    return (not os.path.exists(lock_path), fstat_st)
            finally:
                try:
                    os.unlink(inode_test)
                except OSError as e:
                    if e.errno not in (errno.ENOENT, errno.ESTALE):
                        _raise_exc(e)
            return (True, None)
    finally:
        try:
            os.unlink(hardlink_path)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                _raise_exc(e)
    return (False, fstat_st)


def _fstat_nlink(fd):
    """
    @param fd: an open file descriptor
    @type fd: Integer
    @rtype: Integer
    @return: the current number of hardlinks to the file
    """
    try:
        return os.fstat(fd).st_nlink
    except OSError as e:
        if e.errno in (errno.ENOENT, errno.ESTALE):
            # Some filesystems such as CIFS return
            # ENOENT which means st_nlink == 0.
            return 0
        raise


def unlockfile(mytuple):
    # XXX: Compatability hack.
    if len(mytuple) == 3:
        lockfilename, myfd, unlinkfile = mytuple
        locking_method = fcntl.flock
    elif len(mytuple) == 4:
        lockfilename, myfd, unlinkfile, locking_method = mytuple
    else:
        raise InvalidData

    if myfd == HARDLINK_FD:
        unhardlink_lockfile(lockfilename, unlinkfile=unlinkfile)
        return True

    # myfd may be None here due to myfd = mypath in lockfile()
    if isinstance(lockfilename, str) and not os.path.exists(lockfilename):
        writemsg(_("lockfile does not exist '%s'\n") % lockfilename, 1)
        if myfd is not None:
            _open_fds[myfd].close()
        return False

    try:
        if myfd is None:
            myfd = os.open(lockfilename, os.O_WRONLY, 0o660)
            unlinkfile = 1
        locking_method(myfd, fcntl.LOCK_UN)
    except OSError:
        if isinstance(lockfilename, str):
            _open_fds[myfd].close()
        raise OSError(_("Failed to unlock file '%s'\n") % lockfilename)

    try:
        # This sleep call was added to allow other processes that are
        # waiting for a lock to be able to grab it before it is deleted.
        # lockfile() already accounts for this situation, however, and
        # the sleep here adds more time than is saved overall, so am
        # commenting until it is proved necessary.
        # time.sleep(0.0001)
        if unlinkfile:
            locking_method(myfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            # We won the lock, so there isn't competition for it.
            # We can safely delete the file.
            writemsg(_("Got the lockfile...\n"), 1)
            if _fstat_nlink(myfd) == 1:
                os.unlink(lockfilename)
                writemsg(_("Unlinked lockfile...\n"), 1)
                locking_method(myfd, fcntl.LOCK_UN)
            else:
                writemsg(_("lockfile does not exist '%s'\n") % lockfilename, 1)
                _open_fds[myfd].close()
                return False
    except SystemExit:
        raise
    except Exception as e:
        writemsg(_("Failed to get lock... someone took it.\n"), 1)
        writemsg(str(e) + "\n", 1)

    # why test lockfilename?  because we may have been handed an
    # fd originally, and the caller might not like having their
    # open fd closed automatically on them.
    if isinstance(lockfilename, str):
        _open_fds[myfd].close()

    return True


def hardlock_name(path):
    base, tail = os.path.split(path)
    return os.path.join(
        base,
        ".%s.hardlock-%s-%s"
        % (tail, portage._decode_argv([os.uname()[1]])[0], portage.getpid()),
    )


def hardlink_is_mine(link, lock):
    try:
        lock_st = os.stat(lock)
        if lock_st.st_nlink == 2:
            link_st = os.stat(link)
            return lock_st.st_ino == link_st.st_ino and lock_st.st_dev == link_st.st_dev
    except OSError:
        pass
    return False


def hardlink_lockfile(
    lockfilename, max_wait=DeprecationWarning, waiting_msg=None, flags=0
):
    """Does the NFS, hardlink shuffle to ensure locking on the disk.
    We create a PRIVATE hardlink to the real lockfile, that is just a
    placeholder on the disk.
    If our file can 2 references, then we have the lock. :)
    Otherwise we lather, rise, and repeat.
    """

    if max_wait is not DeprecationWarning:
        warnings.warn(
            "The 'max_wait' parameter of "
            "portage.locks.hardlink_lockfile() is now unused. Use "
            "flags=os.O_NONBLOCK instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    global _quiet
    out = None
    displayed_waiting_msg = False
    preexisting = os.path.exists(lockfilename)
    myhardlock = hardlock_name(lockfilename)

    # Since Python 3.4, chown requires int type (no proxies).
    portage_gid = int(portage.data.portage_gid)

    # myhardlock must not exist prior to our link() call, and we can
    # safely unlink it since its file name is unique to our PID
    try:
        os.unlink(myhardlock)
    except OSError as e:
        if e.errno in (errno.ENOENT, errno.ESTALE):
            pass
        else:
            func_call = f"unlink('{myhardlock}')"
            if e.errno == OperationNotPermitted.errno:
                raise OperationNotPermitted(func_call)
            elif e.errno == PermissionDenied.errno:
                raise PermissionDenied(func_call)
            else:
                raise

    while True:
        # create lockfilename if it doesn't exist yet
        try:
            myfd = os.open(lockfilename, os.O_CREAT | os.O_RDWR, 0o660)
        except OSError as e:
            func_call = f"open('{lockfilename}')"
            if e.errno == OperationNotPermitted.errno:
                raise OperationNotPermitted(func_call)
            elif e.errno == PermissionDenied.errno:
                raise PermissionDenied(func_call)
            elif e.errno == ReadOnlyFileSystem.errno:
                raise ReadOnlyFileSystem(func_call)
            else:
                raise
        else:
            myfd_st = None
            try:
                myfd_st = os.fstat(myfd)
                if not preexisting:
                    # Don't chown the file if it is preexisting, since we
                    # want to preserve existing permissions in that case.
                    if portage.data.secpass >= 1 and myfd_st.st_gid != portage_gid:
                        os.fchown(myfd, -1, portage_gid)
            except OSError as e:
                if e.errno not in (errno.ENOENT, errno.ESTALE):
                    writemsg(
                        "%s: fchown('%s', -1, %d)\n" % (e, lockfilename, portage_gid),
                        noiselevel=-1,
                    )
                    writemsg(
                        _("Cannot chown a lockfile: '%s'\n") % lockfilename,
                        noiselevel=-1,
                    )
                    writemsg(
                        _("Group IDs of current user: %s\n")
                        % " ".join(str(n) for n in os.getgroups()),
                        noiselevel=-1,
                    )
                else:
                    # another process has removed the file, so we'll have
                    # to create it again
                    continue
            finally:
                os.close(myfd)

            # If fstat shows more than one hardlink, then it's extremely
            # unlikely that the following link call will result in a lock,
            # so optimize away the wasteful link call and sleep or raise
            # TryAgain.
            if myfd_st is not None and myfd_st.st_nlink < 2:
                try:
                    os.link(lockfilename, myhardlock)
                except OSError as e:
                    func_call = f"link('{lockfilename}', '{myhardlock}')"
                    if e.errno == OperationNotPermitted.errno:
                        raise OperationNotPermitted(func_call)
                    elif e.errno == PermissionDenied.errno:
                        raise PermissionDenied(func_call)
                    elif e.errno in (errno.ESTALE, errno.ENOENT):
                        # another process has removed the file, so we'll have
                        # to create it again
                        continue
                    else:
                        raise
                else:
                    if hardlink_is_mine(myhardlock, lockfilename):
                        if out is not None:
                            out.eend(os.EX_OK)
                        break

                    try:
                        os.unlink(myhardlock)
                    except OSError as e:
                        # This should not happen, since the file name of
                        # myhardlock is unique to our host and PID,
                        # and the above link() call succeeded.
                        if e.errno not in (errno.ENOENT, errno.ESTALE):
                            raise
                        raise FileNotFound(myhardlock)

        if flags & os.O_NONBLOCK:
            raise TryAgain(lockfilename)

        if out is None and not _quiet:
            out = portage.output.EOutput()
        if out is not None and not displayed_waiting_msg:
            displayed_waiting_msg = True
            if waiting_msg is None:
                waiting_msg = _("waiting for lock on %s\n") % lockfilename
            out.ebegin(waiting_msg)

        time.sleep(_HARDLINK_POLL_LATENCY)

    return True


def unhardlink_lockfile(lockfilename, unlinkfile=True):
    myhardlock = hardlock_name(lockfilename)
    if unlinkfile and hardlink_is_mine(myhardlock, lockfilename):
        # Make sure not to touch lockfilename unless we really have a lock.
        try:
            os.unlink(lockfilename)
        except OSError:
            pass
    try:
        os.unlink(myhardlock)
    except OSError:
        pass


def hardlock_cleanup(path, remove_all_locks=False):
    myhost = portage._decode_argv([os.uname()[1]])[0]
    mydl = os.listdir(path)

    results = []
    mycount = 0

    mylist = {}
    for x in mydl:
        if os.path.isfile(path + "/" + x):
            parts = x.split(".hardlock-")
            if len(parts) == 2:
                filename = parts[0][1:]
                hostpid = parts[1].split("-")
                host = "-".join(hostpid[:-1])
                pid = hostpid[-1]

                if filename not in mylist:
                    mylist[filename] = {}
                if host not in mylist[filename]:
                    mylist[filename][host] = []
                mylist[filename][host].append(pid)

                mycount += 1

    results.append(_("Found %(count)s locks") % {"count": mycount})

    for x in mylist:
        if myhost in mylist[x] or remove_all_locks:
            mylockname = hardlock_name(path + "/" + x)
            if (
                hardlink_is_mine(mylockname, path + "/" + x)
                or not os.path.exists(path + "/" + x)
                or remove_all_locks
            ):
                for y in mylist[x]:
                    for z in mylist[x][y]:
                        filename = path + "/." + x + ".hardlock-" + y + "-" + z
                        if filename == mylockname:
                            continue
                        try:
                            # We're sweeping through, unlinking everyone's locks.
                            os.unlink(filename)
                            results.append(_("Unlinked: ") + filename)
                        except OSError:
                            pass
                try:
                    os.unlink(path + "/" + x)
                    results.append(_("Unlinked: ") + path + "/" + x)
                    os.unlink(mylockname)
                    results.append(_("Unlinked: ") + mylockname)
                except OSError:
                    pass
            else:
                try:
                    os.unlink(mylockname)
                    results.append(_("Unlinked: ") + mylockname)
                except OSError:
                    pass

    return results



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import _emerge.emergelog
from _emerge.EbuildPhase import EbuildPhase
from _emerge.BinpkgFetcher import BinpkgFetcher
from _emerge.BinpkgEnvExtractor import BinpkgEnvExtractor
from _emerge.CompositeTask import CompositeTask
from _emerge.BinpkgVerifier import BinpkgVerifier
from _emerge.EbuildMerge import EbuildMerge
from _emerge.EbuildBuildDir import EbuildBuildDir
from _emerge.SpawnProcess import SpawnProcess
from portage.eapi import eapi_exports_replace_vars
from portage.output import colorize
from portage.util import ensure_dirs
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._dyn_libs.dyn_libs import check_dyn_libs_inconsistent
import portage
from portage import os
from portage import shutil
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
import logging


class Binpkg(CompositeTask):
    __slots__ = (
        "find_blockers",
        "ldpath_mtimes",
        "logger",
        "opts",
        "pkg",
        "pkg_count",
        "prefetcher",
        "settings",
        "world_atom",
    ) + (
        "_bintree",
        "_build_dir",
        "_build_prefix",
        "_ebuild_path",
        "_fetched_pkg",
        "_image_dir",
        "_infloc",
        "_pkg_path",
        "_tree",
        "_verify",
        "_pkg_allocated_path",
    )

    def _writemsg_level(self, msg, level=0, noiselevel=0):
        self.scheduler.output(
            msg,
            level=level,
            noiselevel=noiselevel,
            log_path=self.settings.get("PORTAGE_LOG_FILE"),
        )

    def _start(self):
        pkg = self.pkg
        settings = self.settings
        settings.setcpv(pkg)
        self._tree = "bintree"
        self._bintree = self.pkg.root_config.trees[self._tree]
        self._verify = not self.opts.pretend
        self._pkg_allocated_path = None

        # Use realpath like doebuild_environment() does, since we assert
        # that this path is literally identical to PORTAGE_BUILDDIR.
        dir_path = os.path.join(
            os.path.realpath(settings["PORTAGE_TMPDIR"]),
            "portage",
            pkg.category,
            pkg.pf,
        )
        self._image_dir = os.path.join(dir_path, "image")
        self._infloc = os.path.join(dir_path, "build-info")
        self._ebuild_path = os.path.join(self._infloc, pkg.pf + ".ebuild")
        settings["EBUILD"] = self._ebuild_path
        portage.doebuild_environment(
            self._ebuild_path, "setup", settings=self.settings, db=self._bintree.dbapi
        )
        if dir_path != self.settings["PORTAGE_BUILDDIR"]:
            raise AssertionError(
                f"'{dir_path}' != '{self.settings['PORTAGE_BUILDDIR']}'"
            )
        self._build_dir = EbuildBuildDir(scheduler=self.scheduler, settings=settings)
        settings.configdict["pkg"]["EMERGE_FROM"] = "binary"
        settings.configdict["pkg"]["MERGE_TYPE"] = "binary"

        if eapi_exports_replace_vars(settings["EAPI"]):
            vardb = self.pkg.root_config.trees["vartree"].dbapi
            settings["REPLACING_VERSIONS"] = " ".join(
                {
                    portage.versions.cpv_getversion(x)
                    for x in vardb.match(self.pkg.slot_atom)
                    + vardb.match("=" + self.pkg.cpv)
                }
            )

        # The prefetcher has already completed or it
        # could be running now. If it's running now,
        # wait for it to complete since it holds
        # a lock on the file being fetched. The
        # portage.locks functions are only designed
        # to work between separate processes. Since
        # the lock is held by the current process,
        # use the scheduler and fetcher methods to
        # synchronize with the fetcher.
        prefetcher = self.prefetcher
        if prefetcher is not None and not prefetcher.isAlive():
            # Cancel it because it hasn't started yet.
            prefetcher.cancel()
            self.prefetcher = prefetcher = None
        if prefetcher is None:
            pass
        elif prefetcher.isAlive() and prefetcher.poll() is None:
            if not self.background:
                fetch_log = os.path.join(
                    _emerge.emergelog._emerge_log_dir, "emerge-fetch.log"
                )
                msg = (
                    "Fetching in the background:",
                    prefetcher.pkg_path,
                    "To view fetch progress, run in another terminal:",
                    f"tail -f {fetch_log}",
                )
                out = portage.output.EOutput()
                for l in msg:
                    out.einfo(l)

            self._current_task = prefetcher
            prefetcher.addExitListener(self._prefetch_exit)
            return

        self._prefetch_exit(prefetcher)

    def _prefetch_exit(self, prefetcher):
        if self._was_cancelled():
            self.wait()
            return

        if not (self.opts.pretend or self.opts.fetchonly):
            self._start_task(
                AsyncTaskFuture(future=self._build_dir.async_lock()),
                self._start_fetcher,
            )
        else:
            self._start_fetcher()

    def _start_fetcher(self, lock_task=None):
        if lock_task is not None:
            self._assert_current(lock_task)
            if lock_task.cancelled:
                self._default_final_exit(lock_task)
                return

            lock_task.future.result()
            # Initialize PORTAGE_LOG_FILE (clean_log won't work without it).
            portage.prepare_build_dirs(self.settings["ROOT"], self.settings, 1)
            # If necessary, discard old log so that we don't
            # append to it.
            self._build_dir.clean_log()

        pkg = self.pkg
        pkg_count = self.pkg_count
        fetcher = None

        if self.opts.getbinpkg and self._bintree.isremote(pkg.cpv):
            fetcher = BinpkgFetcher(
                background=self.background,
                logfile=self.settings.get("PORTAGE_LOG_FILE"),
                pkg=self.pkg,
                pretend=self.opts.pretend,
                scheduler=self.scheduler,
            )

            msg = " --- ({} of {}) Fetching Binary ({}::{})".format(
                pkg_count.curval,
                pkg_count.maxval,
                pkg.cpv,
                fetcher.pkg_path,
            )
            short_msg = "emerge: ({} of {}) {} Fetch".format(
                pkg_count.curval,
                pkg_count.maxval,
                pkg.cpv,
            )
            self.logger.log(msg, short_msg=short_msg)

            # Allow the Scheduler's fetch queue to control the
            # number of concurrent fetchers.
            fetcher.addExitListener(self._fetcher_exit)
            self._task_queued(fetcher)
            self.scheduler.fetch.schedule(fetcher)
            return

        self._fetcher_exit(fetcher)

    def _fetcher_exit(self, fetcher):
        # The fetcher only has a returncode when
        # --getbinpkg is enabled.
        if fetcher is not None:
            self._fetched_pkg = fetcher.pkg_path
            self._pkg_allocated_path = fetcher.pkg_allocated_path
            if self._default_exit(fetcher) != os.EX_OK:
                self._async_unlock_builddir(returncode=self.returncode)
                return

        if self.opts.pretend:
            self._current_task = None
            self.returncode = os.EX_OK
            self.wait()
            return

        verifier = None
        if self._verify:
            if self._fetched_pkg:
                path = self._fetched_pkg
            else:
                path = self.pkg.root_config.trees["bintree"].getname(self.pkg.cpv)
            logfile = self.settings.get("PORTAGE_LOG_FILE")
            verifier = BinpkgVerifier(
                background=self.background,
                logfile=logfile,
                pkg=self.pkg,
                scheduler=self.scheduler,
                _pkg_path=path,
            )
            self._start_task(verifier, self._verifier_exit)
            return

        self._verifier_exit(verifier)

    def _verifier_exit(self, verifier):
        if verifier is not None and self._default_exit(verifier) != os.EX_OK:
            self._async_unlock_builddir(returncode=self.returncode)
            return

        logger = self.logger
        pkg = self.pkg
        pkg_count = self.pkg_count

        if self._fetched_pkg:
            pkg_path = self._bintree.getname(
                self._bintree.inject(
                    pkg.cpv,
                    current_pkg_path=self._fetched_pkg,
                    allocated_pkg_path=self._pkg_allocated_path,
                ),
                allocate_new=False,
            )
        else:
            pkg_path = self.pkg.root_config.trees["bintree"].getname(self.pkg.cpv)

        # This gives bashrc users an opportunity to do various things
        # such as remove binary packages after they're installed.
        if pkg_path is not None:
            self.settings["PORTAGE_BINPKG_FILE"] = pkg_path
        self._pkg_path = pkg_path

        logfile = self.settings.get("PORTAGE_LOG_FILE")
        if logfile is not None and os.path.isfile(logfile):
            # Remove fetch log after successful fetch.
            try:
                os.unlink(logfile)
            except OSError:
                pass

        if self.opts.fetchonly:
            self._current_task = None
            self.returncode = os.EX_OK
            self.wait()
            return

        msg = " === ({} of {}) Merging Binary ({}::{})".format(
            pkg_count.curval,
            pkg_count.maxval,
            pkg.cpv,
            pkg_path,
        )
        short_msg = "emerge: ({} of {}) {} Merge Binary".format(
            pkg_count.curval,
            pkg_count.maxval,
            pkg.cpv,
        )
        logger.log(msg, short_msg=short_msg)

        phase = "clean"
        settings = self.settings
        ebuild_phase = EbuildPhase(
            background=self.background,
            phase=phase,
            scheduler=self.scheduler,
            settings=settings,
        )

        self._start_task(ebuild_phase, self._clean_exit)

    def _clean_exit(self, clean_phase):
        if self._default_exit(clean_phase) != os.EX_OK:
            self._async_unlock_builddir(returncode=self.returncode)
            return

        self._start_task(
            AsyncTaskFuture(future=self._unpack_metadata(loop=self.scheduler)),
            self._unpack_metadata_exit,
        )

    async def _unpack_metadata(self, loop=None):
        dir_path = self.settings["PORTAGE_BUILDDIR"]

        infloc = self._infloc
        pkg = self.pkg
        pkg_path = self._pkg_path

        dir_mode = 0o755
        for mydir in (dir_path, self._image_dir, infloc):
            portage.util.ensure_dirs(
                mydir,
                uid=portage.data.portage_uid,
                gid=portage.data.portage_gid,
                mode=dir_mode,
            )

        # This initializes PORTAGE_LOG_FILE.
        portage.prepare_build_dirs(self.settings["ROOT"], self.settings, 1)
        self._writemsg_level(">>> Extracting info\n")

        await self._bintree.dbapi.unpack_metadata(
            self.settings, infloc, loop=self.scheduler
        )
        check_missing_metadata = ("CATEGORY", "PF")
        for k, v in zip(
            check_missing_metadata,
            self._bintree.dbapi.aux_get(self.pkg.cpv, check_missing_metadata),
        ):
            if v:
                continue
            elif k == "CATEGORY":
                v = pkg.category
            elif k == "PF":
                v = pkg.pf
            else:
                continue

            f = open(
                _unicode_encode(
                    os.path.join(infloc, k), encoding=_encodings["fs"], errors="strict"
                ),
                mode="w",
                encoding=_encodings["content"],
                errors="backslashreplace",
            )
            try:
                f.write(_unicode_decode(v + "\n"))
            finally:
                f.close()

        # Store the md5sum in the vdb.
        if pkg_path is not None:
            (md5sum,) = self._bintree.dbapi.aux_get(self.pkg.cpv, ["MD5"])
            if not md5sum:
                md5sum = portage.checksum.perform_md5(pkg_path)
            with open(
                _unicode_encode(
                    os.path.join(infloc, "BINPKGMD5"),
                    encoding=_encodings["fs"],
                    errors="strict",
                ),
                mode="w",
                encoding=_encodings["content"],
                errors="strict",
            ) as f:
                f.write(_unicode_decode(f"{md5sum}\n"))

        env_extractor = BinpkgEnvExtractor(
            background=self.background, scheduler=self.scheduler, settings=self.settings
        )
        env_extractor.start()
        await env_extractor.async_wait()
        if env_extractor.returncode != os.EX_OK:
            raise portage.exception.PortageException(
                f"failed to extract environment for {self.pkg.cpv}"
            )

    def _unpack_metadata_exit(self, unpack_metadata):
        if self._default_exit(unpack_metadata) != os.EX_OK:
            try:
                unpack_metadata.future.result()
            except Exception as e:
                self._writemsg_level(
                    colorize(
                        "BAD",
                        f"!!! Error Extracting '{self._pkg_path}', {e}\n",
                    ),
                    noiselevel=-1,
                    level=logging.ERROR,
                )
            self._async_unlock_builddir(returncode=self.returncode)
            return

        setup_phase = EbuildPhase(
            background=self.background,
            phase="setup",
            scheduler=self.scheduler,
            settings=self.settings,
        )

        setup_phase.addExitListener(self._setup_exit)
        self._task_queued(setup_phase)
        self.scheduler.scheduleSetup(setup_phase)

    def _setup_exit(self, setup_phase):
        if self._default_exit(setup_phase) != os.EX_OK:
            self._async_unlock_builddir(returncode=self.returncode)
            return

        self._writemsg_level(f">>> Extracting {self.pkg.cpv}\n")
        self._start_task(
            AsyncTaskFuture(
                future=self._bintree.dbapi.unpack_contents(
                    self.settings, self._image_dir, loop=self.scheduler
                )
            ),
            self._unpack_contents_exit,
        )

    def _unpack_contents_exit(self, unpack_contents):
        if self._default_exit(unpack_contents) != os.EX_OK:
            try:
                unpack_contents.future.result()
            except Exception as e:
                self._writemsg_level(
                    colorize(
                        "BAD",
                        f"!!! Error Extracting '{self._pkg_path}', {e}\n",
                    ),
                    noiselevel=-1,
                    level=logging.ERROR,
                )
            self._async_unlock_builddir(returncode=self.returncode)
            return

        # Before anything else, let's do an integrity check.
        (provides,) = self._bintree.dbapi.aux_get(self.pkg.cpv, ["PROVIDES"])
        if check_dyn_libs_inconsistent(self.settings["D"], provides):
            self._writemsg_level(
                colorize(
                    "BAD",
                    "!!! Error! Installing dynamic libraries (.so) with blank PROVIDES!",
                ),
                noiselevel=-1,
                level=logging.ERROR,
            )

        try:
            with open(
                _unicode_encode(
                    os.path.join(self._infloc, "EPREFIX"),
                    encoding=_encodings["fs"],
                    errors="strict",
                ),
                encoding=_encodings["repo.content"],
                errors="replace",
            ) as f:
                self._build_prefix = f.read().rstrip("\n")
        except OSError:
            self._build_prefix = ""

        if self._build_prefix == self.settings["EPREFIX"]:
            ensure_dirs(self.settings["ED"])
            self._current_task = None
            self.returncode = os.EX_OK
            self.wait()
            return

        env = self.settings.environ()
        env["PYTHONPATH"] = self.settings["PORTAGE_PYTHONPATH"]
        chpathtool = SpawnProcess(
            args=[
                portage._python_interpreter,
                os.path.join(self.settings["PORTAGE_BIN_PATH"], "chpathtool.py"),
                self.settings["D"],
                self._build_prefix,
                self.settings["EPREFIX"],
            ],
            background=self.background,
            env=env,
            scheduler=self.scheduler,
            logfile=self.settings.get("PORTAGE_LOG_FILE"),
        )
        self._writemsg_level(f">>> Adjusting Prefix to {self.settings['EPREFIX']}\n")
        self._start_task(chpathtool, self._chpathtool_exit)

    def _chpathtool_exit(self, chpathtool):
        if self._final_exit(chpathtool) != os.EX_OK:
            self._writemsg_level(
                f"!!! Error Adjusting Prefix to {self.settings['EPREFIX']}\n",
                noiselevel=-1,
                level=logging.ERROR,
            )
            self._async_unlock_builddir(returncode=self.returncode)
            return

        # We want to install in "our" prefix, not the binary one
        with open(
            _unicode_encode(
                os.path.join(self._infloc, "EPREFIX"),
                encoding=_encodings["fs"],
                errors="strict",
            ),
            mode="w",
            encoding=_encodings["repo.content"],
            errors="strict",
        ) as f:
            f.write(self.settings["EPREFIX"] + "\n")

        # Move the files to the correct location for merge.
        image_tmp_dir = os.path.join(self.settings["PORTAGE_BUILDDIR"], "image_tmp")
        build_d = os.path.join(
            self.settings["D"], self._build_prefix.lstrip(os.sep)
        ).rstrip(os.sep)
        if not os.path.isdir(build_d):
            # Assume this is a virtual package or something.
            shutil.rmtree(self._image_dir)
            ensure_dirs(self.settings["ED"])
        else:
            os.rename(build_d, image_tmp_dir)
            if build_d != self._image_dir:
                shutil.rmtree(self._image_dir)
            ensure_dirs(os.path.dirname(self.settings["ED"].rstrip(os.sep)))
            os.rename(image_tmp_dir, self.settings["ED"])

        self.wait()

    def _async_unlock_builddir(self, returncode=None):
        """
        Release the lock asynchronously, and if a returncode parameter
        is given then set self.returncode and notify exit listeners.
        """
        if self.opts.pretend or self.opts.fetchonly:
            if returncode is not None:
                self.returncode = returncode
                self._async_wait()
            return
        if returncode is not None:
            # The returncode will be set after unlock is complete.
            self.returncode = None
        portage.elog.elog_process(self.pkg.cpv, self.settings)
        self._start_task(
            AsyncTaskFuture(future=self._build_dir.async_unlock()),
            functools.partial(self._unlock_builddir_exit, returncode=returncode),
        )

    def _unlock_builddir_exit(self, unlock_task, returncode=None):
        self._assert_current(unlock_task)
        if unlock_task.cancelled and returncode is not None:
            self._default_final_exit(unlock_task)
            return

        # Normally, async_unlock should not raise an exception here.
        unlock_task.future.cancelled() or unlock_task.future.result()
        if returncode is not None:
            self.returncode = returncode
            self._async_wait()

    def create_install_task(self):
        task = EbuildMerge(
            exit_hook=self._install_exit,
            find_blockers=self.find_blockers,
            ldpath_mtimes=self.ldpath_mtimes,
            logger=self.logger,
            pkg=self.pkg,
            pkg_count=self.pkg_count,
            pkg_path=self._pkg_path,
            scheduler=self.scheduler,
            settings=self.settings,
            tree=self._tree,
            world_atom=self.world_atom,
        )
        return task

    def _install_exit(self, task):
        """
        @returns: Future, result is the returncode from an
                EbuildBuildDir.async_unlock() task
        """
        self.settings.pop("PORTAGE_BINPKG_FILE", None)
        if (
            task.returncode == os.EX_OK
            and "binpkg-logs" not in self.settings.features
            and self.settings.get("PORTAGE_LOG_FILE")
        ):
            try:
                os.unlink(self.settings["PORTAGE_LOG_FILE"])
            except OSError:
                pass
        self._async_unlock_builddir()
        if self._current_task is None:
            result = self.scheduler.create_future()
            self.scheduler.call_soon(result.set_result, os.EX_OK)
        else:
            result = self._current_task.async_wait()
        return result



# Copyright 2011-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class ResolverDepthTestCase(TestCase):
    def testResolverDepth(self):
        profile = {
            "package.mask": (
                # Mask an installed package (for which an update is
                # available) in order to test for bug 712298, where
                # --update caused --deep=<depth> to be violated for
                # such a package.
                "<dev-libs/B-2",
            ),
        }

        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "dev-libs/B"},
            "dev-libs/A-2": {"RDEPEND": "dev-libs/B"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/C"},
            "dev-libs/B-2": {"RDEPEND": "dev-libs/C"},
            "dev-libs/C-1": {},
            "dev-libs/C-2": {},
            "virtual/libusb-0": {
                "EAPI": "2",
                "SLOT": "0",
                "RDEPEND": "|| ( >=dev-libs/libusb-0.1.12-r1:0 dev-libs/libusb-compat >=sys-freebsd/freebsd-lib-8.0[usb] )",
            },
            "virtual/libusb-1": {
                "EAPI": "2",
                "SLOT": "1",
                "RDEPEND": ">=dev-libs/libusb-1.0.4:1",
            },
            "dev-libs/libusb-0.1.13": {},
            "dev-libs/libusb-1.0.5": {"SLOT": "1"},
            "dev-libs/libusb-compat-1": {},
            "sys-freebsd/freebsd-lib-8": {"IUSE": "+usb"},
            "sys-fs/udev-164": {"EAPI": "1", "RDEPEND": "virtual/libusb:0"},
            "virtual/jre-1.5.0": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =dev-java/sun-jre-bin-1.5.0* =virtual/jdk-1.5.0* )",
            },
            "virtual/jre-1.5.0-r1": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =dev-java/sun-jre-bin-1.5.0* =virtual/jdk-1.5.0* )",
            },
            "virtual/jre-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/sun-jre-bin-1.6.0* =virtual/jdk-1.6.0* )",
            },
            "virtual/jre-1.6.0-r1": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/sun-jre-bin-1.6.0* =virtual/jdk-1.6.0* )",
            },
            "virtual/jdk-1.5.0": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =dev-java/sun-jdk-1.5.0* dev-java/gcj-jdk )",
            },
            "virtual/jdk-1.5.0-r1": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =dev-java/sun-jdk-1.5.0* dev-java/gcj-jdk )",
            },
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "virtual/jdk-1.6.0-r1": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "dev-java/gcj-jdk-4.5": {},
            "dev-java/gcj-jdk-4.5-r1": {},
            "dev-java/icedtea-6.1": {},
            "dev-java/icedtea-6.1-r1": {},
            "dev-java/sun-jdk-1.5": {"SLOT": "1.5"},
            "dev-java/sun-jdk-1.6": {"SLOT": "1.6"},
            "dev-java/sun-jre-bin-1.5": {"SLOT": "1.5"},
            "dev-java/sun-jre-bin-1.6": {"SLOT": "1.6"},
            "dev-java/ant-core-1.8": {"DEPEND": ">=virtual/jdk-1.4"},
            "dev-db/hsqldb-1.8": {"RDEPEND": ">=virtual/jre-1.6"},
        }

        installed = {
            "dev-libs/A-1": {"RDEPEND": "dev-libs/B"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/C"},
            "dev-libs/C-1": {},
            "virtual/jre-1.5.0": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =virtual/jdk-1.5.0* =dev-java/sun-jre-bin-1.5.0* )",
            },
            "virtual/jre-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =virtual/jdk-1.6.0* =dev-java/sun-jre-bin-1.6.0* )",
            },
            "virtual/jdk-1.5.0": {
                "SLOT": "1.5",
                "RDEPEND": "|| ( =dev-java/sun-jdk-1.5.0* dev-java/gcj-jdk )",
            },
            "virtual/jdk-1.6.0": {
                "SLOT": "1.6",
                "RDEPEND": "|| ( =dev-java/icedtea-6* =dev-java/sun-jdk-1.6.0* )",
            },
            "dev-java/gcj-jdk-4.5": {},
            "dev-java/icedtea-6.1": {},
            "virtual/libusb-0": {
                "EAPI": "2",
                "SLOT": "0",
                "RDEPEND": "|| ( >=dev-libs/libusb-0.1.12-r1:0 dev-libs/libusb-compat >=sys-freebsd/freebsd-lib-8.0[usb] )",
            },
        }

        world = ["dev-libs/A"]

        test_cases = (
            # Test for bug 712298, where --update caused --deep=<depth>
            # to be violated for dependencies that were masked. In this
            # case, the installed dev-libs/B-1 dependency is masked.
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--update": True, "--deep": 0},
                success=True,
                mergelist=["dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--update": True, "--deep": 1},
                success=True,
                mergelist=["dev-libs/B-2", "dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--update": True, "--deep": 2},
                success=True,
                mergelist=["dev-libs/C-2", "dev-libs/B-2", "dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/C-2", "dev-libs/B-2", "dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--emptytree": True},
                success=True,
                mergelist=["dev-libs/C-2", "dev-libs/B-2", "dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--selective": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--deep": 2},
                success=True,
                mergelist=["dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["virtual/jre"],
                options={},
                success=True,
                mergelist=["virtual/jre-1.6.0-r1"],
            ),
            ResolverPlaygroundTestCase(
                ["virtual/jre"],
                options={"--deep": True},
                success=True,
                mergelist=["virtual/jre-1.6.0-r1"],
            ),
            # Test bug #141118, where we avoid pulling in
            # redundant deps, satisfying nested virtuals
            # as efficiently as possible.
            ResolverPlaygroundTestCase(
                ["virtual/jre"],
                options={"--selective": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
            # Test bug #150361, where depgraph._greedy_slots()
            # is triggered by --update with AtomArg.
            ResolverPlaygroundTestCase(
                ["virtual/jre"],
                options={"--update": True},
                success=True,
                ambiguous_merge_order=True,
                mergelist=[("virtual/jre-1.6.0-r1", "virtual/jre-1.5.0-r1")],
            ),
            # Recursively traversed virtual dependencies, and their
            # direct dependencies, are considered to have the same
            # depth as direct dependencies.
            ResolverPlaygroundTestCase(
                ["virtual/jre"],
                options={"--update": True, "--deep": 1},
                success=True,
                ambiguous_merge_order=True,
                merge_order_assertions=(
                    ("dev-java/icedtea-6.1-r1", "virtual/jdk-1.6.0-r1"),
                    ("virtual/jdk-1.6.0-r1", "virtual/jre-1.6.0-r1"),
                    ("dev-java/gcj-jdk-4.5-r1", "virtual/jdk-1.5.0-r1"),
                    ("virtual/jdk-1.5.0-r1", "virtual/jre-1.5.0-r1"),
                ),
                mergelist=[
                    (
                        "dev-java/icedtea-6.1-r1",
                        "dev-java/gcj-jdk-4.5-r1",
                        "virtual/jdk-1.6.0-r1",
                        "virtual/jdk-1.5.0-r1",
                        "virtual/jre-1.6.0-r1",
                        "virtual/jre-1.5.0-r1",
                    )
                ],
            ),
            ResolverPlaygroundTestCase(
                ["virtual/jre:1.5"],
                options={"--update": True},
                success=True,
                mergelist=["virtual/jre-1.5.0-r1"],
            ),
            ResolverPlaygroundTestCase(
                ["virtual/jre:1.6"],
                options={"--update": True},
                success=True,
                mergelist=["virtual/jre-1.6.0-r1"],
            ),
            # Test that we don't pull in any unnecessary updates
            # when --update is not specified, even though we
            # specified --deep.
            ResolverPlaygroundTestCase(
                ["dev-java/ant-core"],
                options={"--deep": True},
                success=True,
                mergelist=["dev-java/ant-core-1.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-java/ant-core"],
                options={"--update": True},
                success=True,
                mergelist=["dev-java/ant-core-1.8"],
            ),
            # Recursively traversed virtual dependencies, and their
            # direct dependencies, are considered to have the same
            # depth as direct dependencies.
            ResolverPlaygroundTestCase(
                ["dev-java/ant-core"],
                options={"--update": True, "--deep": 1},
                success=True,
                mergelist=[
                    "dev-java/icedtea-6.1-r1",
                    "virtual/jdk-1.6.0-r1",
                    "dev-java/ant-core-1.8",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["dev-db/hsqldb"],
                options={"--deep": True},
                success=True,
                mergelist=["dev-db/hsqldb-1.8"],
            ),
            # Don't traverse deps of an installed package with --deep=0,
            # even if it's a virtual.
            ResolverPlaygroundTestCase(
                ["virtual/libusb:0"],
                options={"--selective": True, "--deep": 0},
                success=True,
                mergelist=[],
            ),
            # Satisfy unsatisfied dep of installed package with --deep=1.
            ResolverPlaygroundTestCase(
                ["virtual/libusb:0"],
                options={"--selective": True, "--deep": 1},
                success=True,
                mergelist=["dev-libs/libusb-0.1.13"],
            ),
            # Pull in direct dep of virtual, even with --deep=0.
            ResolverPlaygroundTestCase(
                ["sys-fs/udev"],
                options={"--deep": 0},
                success=True,
                mergelist=["dev-libs/libusb-0.1.13", "sys-fs/udev-164"],
            ),
            # Test --nodeps with direct virtual deps.
            ResolverPlaygroundTestCase(
                ["sys-fs/udev"],
                options={"--nodeps": True},
                success=True,
                mergelist=["sys-fs/udev-164"],
            ),
            # Test that --nodeps overrides --deep.
            ResolverPlaygroundTestCase(
                ["sys-fs/udev"],
                options={"--nodeps": True, "--deep": True},
                success=True,
                mergelist=["sys-fs/udev-164"],
            ),
            # Test that --nodeps overrides --emptytree.
            ResolverPlaygroundTestCase(
                ["sys-fs/udev"],
                options={"--nodeps": True, "--emptytree": True},
                success=True,
                mergelist=["sys-fs/udev-164"],
            ),
            # Test --emptytree with virtuals.
            ResolverPlaygroundTestCase(
                ["sys-fs/udev"],
                options={"--emptytree": True},
                success=True,
                mergelist=[
                    "dev-libs/libusb-0.1.13",
                    "virtual/libusb-0",
                    "sys-fs/udev-164",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, profile=profile, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()
